from memalot.base import MemalotCount
from memalot.memory import (
    MemalotMemoryUsage,
    get_memory_usage,
)


class TestGetMemoryUsage:
    """
    Tests the `get_memory_usage` function.
    """

    def test_get_memory_usage(self) -> None:
        """
        Test get_memory_usage.
        """
        result = get_memory_usage(MemalotCount(11))

        # There isn't much we can check here, but we can at least make sure
        # the result is a MemoryUsage object and has the expected fields.
        assert isinstance(result, MemalotMemoryUsage)
        assert isinstance(result.current_rss_bytes, int)
        assert result.current_rss_bytes > 0
        assert isinstance(result.system_percent_used, float)
        assert 0 <= result.system_percent_used <= 100
        assert isinstance(result.peak_rss_bytes, int)
        assert result.peak_rss_bytes > 0
        assert result.iteration_number == 11



from types import CellType
from typing import AbstractSet, Any, Type

from memalot.base import MemalotObjectIds
from memalot.utils import get_full_type_name, get_objects


def _is_included_type(
    obj: Any,
    excluded_types: set[Type[Any]],
    included_type_names: AbstractSet[str],
    excluded_type_names: AbstractSet[str],
    include_object_ids: MemalotObjectIds,
    exclude_object_ids: MemalotObjectIds,
) -> bool:
    """
    Returns `True` if the object should be included, based on the inclusion and exclusion rules.
    """
    if (
        id(obj) == id(excluded_types)
        or id(obj) == id(included_type_names)
        or id(obj) == id(excluded_type_names)
        or id(obj) == id(include_object_ids)
        or id(obj) == id(exclude_object_ids)
        or id(obj) in exclude_object_ids
    ):
        return False
    if include_object_ids and id(obj) not in include_object_ids:
        return False
    if type(obj) in excluded_types:
        return False
    # Also exclude objects where an excluded type is referred to by a cell (closure)
    if isinstance(obj, CellType):
        try:
            if type(obj.cell_contents) in excluded_types:
                return False
        except Exception:  # pragma: nocover
            # If we can't get the contents of the cell, just carry on.
            pass
    type_name = get_full_type_name(type(obj))
    if included_type_names and not any(
        included_type in type_name for included_type in included_type_names
    ):
        return False
    if excluded_type_names and any(
        excluded_type in type_name for excluded_type in excluded_type_names
    ):
        return False
    return True


def filter_objects(
    objects: list[Any],
    excluded_types: set[Type[Any]],
    included_type_names: AbstractSet[str],
    excluded_type_names: AbstractSet[str],
    include_object_ids: MemalotObjectIds,
    exclude_object_ids: MemalotObjectIds,
) -> list[Any]:
    """
    Filters the provided `objects` according to a set of rules.

    Objects are returned in the same order they are passed in.

    :param objects: The objects to filter.
    :param excluded_types: Concrete Python types to exclude from the results.
    :param included_type_names: Substrings of fully-qualified type names that must match for
        objects to be included (when non-empty).
    :param excluded_type_names: Substrings of fully-qualified type names that cause objects to
        be excluded when any match.
    :param include_object_ids: If non-empty, only objects whose IDs are contained here will be
        included in the result.
    :param exclude_object_ids: Objects whose IDs are contained here will be excluded.
    :return: A new list containing the objects that satisfy the inclusion rules.
    """
    # Also exclude the IDs of the objects in the include_object_ids and exclude_object_ids lists
    exclude_object_ids = MemalotObjectIds(
        exclude_object_ids
        | {id(include_object_id) for include_object_id in include_object_ids}
        | {id(exclude_object_id) for exclude_object_id in exclude_object_ids}
    )
    exclude_dict_ids = set()
    return_objects = []
    for obj in objects:
        if _is_included_type(
            obj=obj,
            excluded_types=excluded_types,
            included_type_names=included_type_names,
            excluded_type_names=excluded_type_names,
            include_object_ids=include_object_ids,
            exclude_object_ids=exclude_object_ids,
        ):
            return_objects.append(obj)
        if _safe_hasattr(obj, "__dict__"):
            exclude_dict_ids.add(id(obj.__dict__))
    # Do another pass to exclude the dicts of the objects. This is necessary in Python versions
    # <3.12 because gc.get_objects() also returns the dicts of objects
    # as well as the objects themselves, which adds noise to the results.
    return [obj for obj in return_objects if id(obj) not in exclude_dict_ids]


def gc_and_get_objects(max_untracked_search_depth: int) -> list[Any]:
    """
    Performs a garbage collection and returns the in-memory, non-collectable objects.

    :param max_untracked_search_depth: The maximum depth to search for untracked objects.
    :return: A list of live objects currently reachable in the process.
    """
    # The utils.get_objects function takes care of garbage collection
    return get_objects(max_untracked_search_depth=max_untracked_search_depth)


def _safe_hasattr(obj: Any, attr_name: str) -> bool:
    try:
        return hasattr(obj, attr_name)
    except Exception:  # pragma: nocover
        # Some things don't like hasattr being called on them.
        return False



from memalot.api import (
    create_leak_monitor as create_leak_monitor,
)
from memalot.api import (
    leak_monitor as leak_monitor,
)
from memalot.api import (
    start_leak_monitoring as start_leak_monitoring,
)
from memalot.interface import LeakMonitor as LeakMonitor
from memalot.interface import Stoppable as Stoppable
from memalot.options import Options as Options



import sys
from abc import ABC, abstractmethod
from collections import Counter, defaultdict
from datetime import datetime, timezone
from typing import Any, Callable, Dict, Generator, Tuple, Type
from weakref import WeakValueDictionary

import objsize
import referrers
from referrers import ReferrerGraph

from memalot.base import (
    ApproximateSize,
    MemalotCount,
    MemalotObjectId,
    MemalotObjectIds,
    MemalotSet,
    ObjectGetter,
    ObjectSignature,
)
from memalot.memory import (
    MemalotMemoryUsage,
    get_memory_usage,
)
from memalot.options import Options
from memalot.reports import (
    LeakSummary,
    ObjectDetails,
    TypeSummary,
)
from memalot.utils import (
    convert_graph_nodes,
    get_full_type_name,
    get_module_prefix,
    get_object_type_budgets,
)

# Names to replace in the referrer graph when using a decorator. This is obviously a
# bit brittle, but hopefully integration tests will catch any issues.
_DECORATOR_WRAPPER_NAME = "memalot_decorator_inner_wrapper"
_DECORATOR_ARGS_NAME = "memalot_decorator_inner_args"
_DECORATOR_KWARGS_NAME = "memalot_decorator_inner_kwargs"
_ARGS_SUBSTITUTION = f"{_DECORATOR_WRAPPER_NAME}.{_DECORATOR_ARGS_NAME}"
_KWARGS_SUBSTITUTION = f"{_DECORATOR_WRAPPER_NAME}.{_DECORATOR_KWARGS_NAME}"


class MemalotObjects:
    """
    A collection of objects.

    A summary can be produced via the `get_leak_summary` method, and object details can be
    obtained via the `generate_object_details` method.
    """

    def __init__(self, objects: list[Any]) -> None:
        self._objects = objects

    def get_leak_summary(self, iteration: MemalotCount, options: Options) -> LeakSummary:
        """
        Gets a summary of potential leaks.
        """
        type_counts = Counter(type(o) for o in self._objects)
        type_sizes: dict[Type[Any], ApproximateSize] = defaultdict(ApproximateSize)
        for obj in self._objects:
            obj_type = type(obj)
            if options.compute_size_in_leak_summary:
                type_sizes[obj_type] += _safe_shallow_size(obj)
        type_summaries = []
        for obj_type, count in type_counts.most_common(options.max_types_in_leak_summary):
            if options.compute_size_in_leak_summary:
                size = type_sizes[obj_type]
            else:
                size = None
            type_summaries.append(
                TypeSummary(
                    object_type=get_full_type_name(obj_type),
                    count=count,
                    shallow_size_bytes=size,
                )
            )
        return LeakSummary(
            iteration=int(iteration),
            type_summaries=type_summaries,
            max_types_in_summary=options.max_types_in_leak_summary,
        )

    def generate_object_details(
        self,
        excluded_from_referrers: list[int],
        options: Options,
        function_name: str | None = None,
        get_referrers_func: Callable[..., ReferrerGraph] = referrers.get_referrer_graph,
    ) -> Generator[ObjectDetails, None, None]:
        if not self._objects:
            return

        # Figure out how many objects to return per type based on options.max_referrers.
        # If we have more types than options.max_referrers, then we return one of each
        # of the most common types.
        type_counts = Counter(type(o) for o in self._objects)
        most_common_types = dict(type_counts.most_common(options.max_object_details))
        type_budgets = get_object_type_budgets(
            object_type_counts=most_common_types, total_budget=options.max_object_details
        )

        # Replace decorator wrapper names with the decorated function name.
        replacements = {}
        if function_name is not None:
            replacements[_ARGS_SUBSTITUTION] = f"{function_name} args"
            replacements[_KWARGS_SUBSTITUTION] = f"{function_name} kwargs"

        # If module prefixes are not specified, we use the module prefix of the caller
        # (not the direct caller of this method; the caller of Memalot)
        if options.referrers_module_prefixes is not None:
            module_prefixes = options.referrers_module_prefixes
        else:
            prefix = get_module_prefix()
            if prefix is not None:
                module_prefixes = {prefix}
            else:  # pragma: no cover
                module_prefixes = None

        # Return objects in order of the most common types.
        sorted_objects = sorted(self._objects, key=lambda x: type_counts[type(x)], reverse=True)
        # Get an iterator rather than using a for loop, so we can exclude it from the
        # referrers.
        iterator = iter(sorted_objects)
        current_counts_per_type: Counter[type] = Counter()
        str_func = options.str_func if options.str_func else self._safe_str
        try:
            while True:
                # Wrap the object in a list so we can exclude it
                obj_list = [next(iterator)]
                obj_type = type(obj_list[0])
                current_counts_per_type[obj_type] += 1
                if (
                    obj_type not in most_common_types
                    or current_counts_per_type[obj_type] > type_budgets[obj_type]
                ):
                    continue
                if options.check_referrers:
                    referrer_graph = get_referrers_func(
                        obj_list[0],
                        exclude_object_ids=[
                            id(sorted_objects),
                            id(iterator),
                            id(self._objects),
                            id(obj_list),
                        ]
                        + excluded_from_referrers,
                        max_depth=options.referrers_max_depth,
                        max_untracked_search_depth=options.referrers_max_untracked_search_depth,
                        timeout=options.referrers_search_timeout,
                        single_object_referrer_limit=options.single_object_referrer_limit,
                        module_prefixes=module_prefixes,
                    )
                else:
                    referrer_graph = None
                yield ObjectDetails(
                    deep_size_bytes=_safe_deep_size(obj_list[0]),
                    object_type_name=get_full_type_name(type(obj_list[0])),
                    object_id=id(obj_list[0]),
                    object_str=str_func(obj_list[0], options.str_max_length),
                    referrer_graph=(
                        convert_graph_nodes(
                            referrer_graph.to_networkx(),
                            replacements=replacements,
                        )
                        if referrer_graph
                        else None
                    ),
                    referrers_checked=options.check_referrers,
                )
        except StopIteration:
            pass

    def __len__(self) -> int:
        return len(self._objects)

    def _safe_str(self, obj: Any, truncate_at: int) -> str:
        try:
            str_repr = str(obj)
            if len(str_repr) > truncate_at:  # pragma: no cover
                str_repr = str_repr[:truncate_at] + f" … ({len(str_repr) - truncate_at} more chars)"
            return str_repr
        except Exception as e:
            # Some things don't like their string representation being obtained.
            return f"<Error when getting string representation: {str(e)}>"

    @property
    def objects(self) -> list[Any]:
        return self._objects

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MemalotObjects):
            return False
        return self._objects == other._objects


class MemalotUsageSnapshot:
    """
    A snapshot of memory usage at a particular point in time.
    """

    def __init__(
        self,
        object_getter: ObjectGetter,
        filter_condition: Callable[[Any], bool] | None,
    ) -> None:
        objects = object_getter.get_objects()
        # We use a WeakValueDictionary here to avoid leaking memory. If an object disappears
        # from the dict then it has been garbage collected. If we later find an object with
        # the same ID, that is a different object.
        # However, some builtins (lists, dicts, etc.) do not support weak refs. For these,
        # we use ObjectSignature which allows objects to be compared using heuristics.
        self._weak_ref_objects: WeakValueDictionary[MemalotObjectId, Any] = WeakValueDictionary()
        self._non_weak_ref_objects: Dict[MemalotObjectId, ObjectSignature] = {}
        for obj in [obj for obj in objects if filter_condition is None or filter_condition(obj)]:
            object_id = MemalotObjectId(id(obj))
            try:
                self._weak_ref_objects[object_id] = obj
            except TypeError:
                # Some built-in objects do not support weak-refs (lists, dicts, etc).
                self._non_weak_ref_objects[object_id] = ObjectSignature(obj)
        # Get objects that were created during snapshot creation. These need to be excluded.
        # We just use a combination of object ID and type here to avoid creating more objects.
        # This isn't perfect, but it's probably good enough, particularly since most objects
        # created during snapshot creation will live as long as the snapshot itself.
        # Note: this will exclude some objects that were created in other threads during
        # snapshot creation, but we don't have a better solution for now, so this is called
        # out in the limitations section of the docs.
        (
            self._ids_during_snapshot_creation,
            self._types_during_snapshot_creation,
        ) = self._get_added_during_snapshot_creation(
            objects_before_snapshot_creation=objects,
            object_getter=object_getter,
        )

    def is_new_since_snapshot(self, obj: Any) -> bool:
        """
        Returns `True` if the given object is (probably) new since the snapshot was taken.

        We know for sure whether weakly referenceable objects were part of the snapshot. For
        non-weakly referenceable objects, we base this on heuristics, so we do not know for sure.
        """
        object_id = MemalotObjectId(id(obj))
        if (
            object_id in self._ids_during_snapshot_creation
            and type(obj) in self._types_during_snapshot_creation
        ):
            return False
        if object_id in self._weak_ref_objects:
            return False
        elif object_id in self._non_weak_ref_objects:
            return not self._non_weak_ref_objects[object_id].is_probably_same_object(obj)
        else:
            return True

    def is_in_snapshot(self, obj: Any) -> bool:
        """
        Returns `True` if the given object is in the snapshot.
        """
        object_id = MemalotObjectId(id(obj))
        return object_id in self._weak_ref_objects or (
            object_id in self._non_weak_ref_objects
            and self._non_weak_ref_objects[object_id].is_probably_same_object(obj)
        )

    def _get_added_during_snapshot_creation(
        self,
        objects_before_snapshot_creation: list[Any],
        object_getter: ObjectGetter,
    ) -> Tuple[MemalotObjectIds, MemalotSet]:
        """
        Gets object IDs of objects that were created during snapshot creation. These are internal
        machinery and need to be excluded.
        """
        # Note: there should not be any issue with using object IDs here, since the objects
        # in `objects_before_snapshot_creation` remain in scope while this function is called
        # and so the IDs of the objects in that list cannot be reused (they are not eligible
        # for garbage collection).
        id_set = MemalotObjectIds()
        type_set = MemalotSet()
        obj_object_ids = {id(obj) for obj in objects_before_snapshot_creation}
        for new_obj in object_getter.get_objects():
            new_obj_id = id(new_obj)
            if new_obj_id not in obj_object_ids:
                id_set.add(MemalotObjectId(new_obj_id))
                type_set.add(type(new_obj))
        id_set_id = MemalotObjectId(id(id_set))
        id_set.add(id_set_id)
        type_set.add(type(id_set_id))
        type_set_id = MemalotObjectId(id(type_set))
        id_set.add(type_set_id)
        type_set.add(type(type_set_id))
        return id_set, type_set

    @property
    def active_object_ids(self) -> MemalotObjectIds:
        """
        Returns the IDs of all objects in the snapshot. This includes objects that are
        weakly-referenceable and have not been garbage collected, and *all* objects that are
        not weakly-referenceable.

        Do not use this to determine whether an object is in the snapshot or not - use
        `is_new_since_snapshot()` for that.
        """
        return MemalotObjectIds(
            set(self._weak_ref_objects.keys()) | set(self._non_weak_ref_objects.keys())
        )


class MemoryUsageProvider(ABC):
    @abstractmethod
    def rotate_memory_usage(  # pragma: no cover
        self, iteration: MemalotCount
    ) -> Tuple[MemalotMemoryUsage | None, MemalotMemoryUsage]:
        pass


class MemalotSnapshotManager(MemoryUsageProvider):
    """
    Manages one or more memory snapshots and allows a report to be generated based on
    these.
    """

    def __init__(
        self,
        report_id: str,
    ) -> None:
        self._report_id = report_id

        # Mutable state
        self._most_recent_snapshot: MemalotUsageSnapshot | None = None
        self._most_recent_snapshot_time: datetime | None = None
        self._added_during_snapshot_creation: MemalotObjectIds | None = None
        self._most_recent_reported_usage: MemalotMemoryUsage | None = None

    def generate_new_snapshot(
        self,
        object_getter: ObjectGetter,
        since_snapshot: MemalotUsageSnapshot | None,
    ) -> None:
        if since_snapshot:
            filter_condition = since_snapshot.is_new_since_snapshot
        else:
            filter_condition = None
        self._most_recent_snapshot = MemalotUsageSnapshot(
            object_getter=object_getter, filter_condition=filter_condition
        )
        self._most_recent_snapshot_time = datetime.now(timezone.utc)

    def clear_snapshots(self) -> None:
        self._most_recent_snapshot = None
        self._most_recent_snapshot_time = None

    def rotate_memory_usage(
        self, iteration: MemalotCount
    ) -> Tuple[MemalotMemoryUsage | None, MemalotMemoryUsage]:
        old_usage = self._most_recent_reported_usage
        self._most_recent_reported_usage = get_memory_usage(iteration=iteration)
        return old_usage, self._most_recent_reported_usage

    @property
    def report_id(self) -> str:
        return self._report_id

    @property
    def most_recent_snapshot(self) -> MemalotUsageSnapshot | None:
        return self._most_recent_snapshot

    @property
    def most_recent_snapshot_time(self) -> datetime | None:
        return self._most_recent_snapshot_time

    @property
    def most_recent_reported_usage(self) -> MemalotMemoryUsage | None:
        return self._most_recent_reported_usage


def _safe_deep_size(obj: Any) -> ApproximateSize:
    """
    Gets the approximate deep size of an object. If an error is encountered getting
    the deep size, then an `ApproximateSize` where the upper bound is unknown is returned.
    """
    try:
        return ApproximateSize(
            approx_size=objsize.get_deep_size(obj),
        )
    except Exception:  # pragma: no cover
        return ApproximateSize(approx_size=0, upper_bound_known=False)


def _safe_shallow_size(obj: Any) -> ApproximateSize:
    """
    Gets the approximate shallow size of an object. If an error is encountered getting
    the deep size, then an `ApproximateSize` where the upper bound is unknown is returned.
    """
    try:
        return ApproximateSize(
            approx_size=sys.getsizeof(obj),
        )
    except Exception:  # pragma: no cover
        return ApproximateSize(approx_size=0, upper_bound_known=False)



from typing import Any, Callable, Type

import pytest

from memalot.base import MemalotObjectIds
from memalot.objects import filter_objects


class _CustomA:
    pass


class _CustomB:
    pass


customa1 = _CustomA()
customa2 = _CustomA()
customb1 = _CustomB()
string1 = "string"
int1 = 123
the_list: list[Any] = [1, 2, 3]
the_dict: dict[str, Any] = {"k": "v"}

sample_objects = [
    customa1,
    customa2,
    customb1,
    string1,
    int1,
    the_list,
    the_dict,
]


class TestFilterObjects:
    """
    Tests for the `filter_objects` function.
    """

    @pytest.mark.parametrize(
        "included_type_names,excluded_type_names,include_ids,exclude_ids,expected",
        [
            pytest.param(
                set(),
                set(),
                set(),
                set(),
                [customa1, customa2, customb1, string1, the_list, the_dict],
                id="none",
            ),
            pytest.param(
                {"Custom"},
                set(),
                set(),
                set(),
                [customa1, customa2, customb1],
                id="only_included_type_names",
            ),
            pytest.param(
                set(),
                {"Custom"},
                set(),
                set(),
                [string1, the_list, the_dict],
                id="only_excluded_type_names",
            ),
            pytest.param(
                {"Custom"},
                {"CustomB"},
                set(),
                set(),
                [customa1, customa2],
                id="include_and_exclude_names",
            ),
            pytest.param(
                set(),
                set(),
                {id(customa1), id(the_list)},
                set(),
                [customa1, the_list],
                id="include_object_ids",
            ),
            pytest.param(
                set(),
                set(),
                set(),
                {id(customa2), id(the_dict)},
                [customa1, customb1, string1, the_list],
                id="exclude_object_ids",
            ),
            pytest.param(
                {"Custom"},
                set(),
                set(),
                {id(customa2)},
                [customa1, customb1],
                id="names_and_exclude_ids",
            ),
            pytest.param(
                {"Custom"},
                set(),
                {id(customb1)},
                set(),
                [customb1],
                id="names_and_include_ids_subselection",
            ),
            pytest.param(
                {"Custom"},
                set(),
                {id(the_dict)},
                set(),
                [],
                id="include_ids_conflict_with_names",
            ),
            pytest.param(
                {"DoesNotExist"},
                set(),
                set(),
                set(),
                [],
                id="included_names_no_results",
            ),
            pytest.param(
                set(),
                {"NoSuch"},
                set(),
                set(),
                [customa1, customa2, customb1, string1, the_list, the_dict],
                id="excluded_names_no_effect",
            ),
            pytest.param(
                {"Custom"},
                {"CustomB"},
                {id(customa1), id(the_list)},
                {id(customa2), id(the_dict)},
                [customa1],
                id="combined",
            ),
        ],
    )
    def test_filter_objects_parametrized(
        self,
        included_type_names: set[str],
        excluded_type_names: set[str],
        include_ids: set[int],
        exclude_ids: set[int],
        expected: list[Any],
    ) -> None:
        """
        Tests the filter_objects function with various inclusions and exclusions.
        """
        excluded_types: set[Type[Any]] = {int}
        result = filter_objects(
            objects=sample_objects,
            excluded_types=excluded_types,
            included_type_names=included_type_names,
            excluded_type_names=excluded_type_names,
            include_object_ids=MemalotObjectIds(include_ids),
            exclude_object_ids=MemalotObjectIds(exclude_ids),
        )

        assert result == expected

    def test_closure_filtering_with_excluded_type(self) -> None:
        """
        Tests that closure cells and their contents are properly filtered when the
        type is excluded.
        """
        captured_object = _CustomA()

        # Create a closure that captures the _CustomA object
        def create_closure() -> Callable[[], _CustomA]:
            def inner_function() -> _CustomA:
                return captured_object

            return inner_function

        closure = create_closure()

        # Get the closure cell directly from the closure
        assert closure.__closure__ is not None
        closure_cell = closure.__closure__[0]

        test_objects = [captured_object, closure_cell]

        # When _CustomA is not excluded, both cell and object should be present
        result_not_excluded = filter_objects(
            objects=test_objects,
            excluded_types=set(),  # Don't exclude _CustomA
            included_type_names=set(),
            excluded_type_names=set(),
            include_object_ids=MemalotObjectIds(),
            exclude_object_ids=MemalotObjectIds(),
        )

        # Both the closure cell and the captured object should be present
        assert captured_object in result_not_excluded
        assert closure_cell in result_not_excluded

        # When _CustomA is excluded, both cell and object should be absent
        result_excluded = filter_objects(
            objects=test_objects,
            excluded_types={_CustomA},  # Exclude _CustomA
            included_type_names=set(),
            excluded_type_names=set(),
            include_object_ids=MemalotObjectIds(),
            exclude_object_ids=MemalotObjectIds(),
        )

        # Neither the closure cell nor the captured object should be present
        assert captured_object not in result_excluded
        assert closure_cell not in result_excluded


def test_gc_and_get_objects() -> None:
    """
    Test that gc_and_get_objects returns a list of objects.

    This is just a smoke test. Most testing of the underlying functionality is done in
    the tests for `get_objects`.
    """
    from memalot.objects import gc_and_get_objects

    result = gc_and_get_objects(max_untracked_search_depth=1)

    assert isinstance(result, list)
    assert len(result) > 0  # Should have some objects
    # Check that we get various types of objects
    types_found = {type(obj) for obj in result}
    assert len(types_found) > 1  # Should find multiple types



from datetime import datetime, timezone
from pathlib import Path
from typing import Any

import pytest
from fastmcp import Client
from fastmcp.exceptions import ToolError

from memalot.mcp_server.main import mcp


@pytest.fixture(name="test_reports_dir")
def _test_reports_dir() -> Path:
    """
    Fixture providing the path to test reports directory.
    """
    return Path(__file__).parent.parent / "test_data" / "reports" / "v1"


@pytest.fixture(name="mcp_client")
def _mcp_client() -> "Client[Any]":
    """
    Fixture providing a FastMCP client connected to the server.
    """
    return Client[Any](mcp)


class TestListLeakReports:
    """
    Tests for the `list_leak_reports` MCP tool.
    """

    @pytest.mark.asyncio
    async def test_list_reports_default_parameters(
        self, mcp_client: Client[Any], test_reports_dir: Path
    ) -> None:
        """
        Test listing reports with default parameters (5 reports, default directory).
        """
        async with mcp_client:
            result = await mcp_client.call_tool(
                "list_leak_reports", {"report_directory": str(test_reports_dir)}
            )

        assert hasattr(result, "data"), f"Result {type(result)} has no data attribute"
        assert not result.is_error, "Tool call should not result in an error"
        report_summaries = result.data

        assert len(report_summaries.summaries) == 3
        assert report_summaries.summaries[0].metadata.report_id == "d6wf-uw2w"
        assert report_summaries.summaries[0].metadata.entrypoint == "utils_for_testing.py"
        assert report_summaries.summaries[0].metadata.arguments == []
        assert report_summaries.summaries[0].metadata.start_time == datetime(
            2025, 9, 29, 20, 54, 18, 936290, tzinfo=timezone.utc
        )
        assert report_summaries.summaries[0].iteration_count == 3

        assert report_summaries.summaries[1].metadata.report_id == "lyy0-uf3p"
        assert report_summaries.summaries[1].metadata.entrypoint == "utils_for_testing.py"
        assert report_summaries.summaries[1].metadata.arguments == []
        assert report_summaries.summaries[1].metadata.start_time == datetime(
            2025, 9, 29, 20, 54, 6, 560032, tzinfo=timezone.utc
        )
        assert report_summaries.summaries[1].iteration_count == 3

        assert report_summaries.summaries[2].metadata.report_id == "ykik-ab4l"
        assert report_summaries.summaries[2].metadata.entrypoint == "utils_for_testing.py"
        assert report_summaries.summaries[2].metadata.arguments == ["dummy_arg1", "dummy_arg2"]
        assert report_summaries.summaries[2].metadata.start_time == datetime(
            2025, 9, 29, 20, 53, 39, 600135, tzinfo=timezone.utc
        )
        assert report_summaries.summaries[2].iteration_count == 3

    @pytest.mark.asyncio
    async def test_list_reports_with_num_reports(
        self, mcp_client: Client[Any], test_reports_dir: Path
    ) -> None:
        """
        Test listing reports with specific number of reports requested.
        """
        async with mcp_client:
            result = await mcp_client.call_tool(
                "list_leak_reports", {"num_reports": 2, "report_directory": str(test_reports_dir)}
            )

        assert hasattr(result, "data"), f"Result {type(result)} has no data attribute"
        assert not result.is_error, "Tool call should not result in an error"
        report_summaries = result.data

        assert len(report_summaries.summaries) == 2
        assert report_summaries.summaries[0].metadata.report_id == "d6wf-uw2w"
        assert report_summaries.summaries[0].metadata.entrypoint == "utils_for_testing.py"
        assert report_summaries.summaries[0].metadata.arguments == []
        assert report_summaries.summaries[0].metadata.start_time == datetime(
            2025, 9, 29, 20, 54, 18, 936290, tzinfo=timezone.utc
        )
        assert report_summaries.summaries[0].iteration_count == 3

        assert report_summaries.summaries[1].metadata.report_id == "lyy0-uf3p"
        assert report_summaries.summaries[1].metadata.entrypoint == "utils_for_testing.py"
        assert report_summaries.summaries[1].metadata.arguments == []
        assert report_summaries.summaries[1].metadata.start_time == datetime(
            2025, 9, 29, 20, 54, 6, 560032, tzinfo=timezone.utc
        )
        assert report_summaries.summaries[1].iteration_count == 3

    @pytest.mark.asyncio
    async def test_list_reports_high_num_reports(
        self, mcp_client: Client[Any], test_reports_dir: Path
    ) -> None:
        """
        Test listing reports when requesting more reports than available.
        """
        async with mcp_client:
            result = await mcp_client.call_tool(
                "list_leak_reports", {"num_reports": 100, "report_directory": str(test_reports_dir)}
            )

        assert hasattr(result, "data"), f"Result {type(result)} has no data attribute"
        assert not result.is_error, "Tool call should not result in an error"
        report_summaries = result.data
        report_ids = [report.metadata.report_id for report in report_summaries.summaries]
        assert "d6wf-uw2w" in report_ids
        assert "lyy0-uf3p" in report_ids
        assert "ykik-ab4l" in report_ids

    @pytest.mark.asyncio
    async def test_list_reports_nonexistent_directory(
        self, mcp_client: Client[Any], tmp_path: Path
    ) -> None:
        """
        Test listing reports from a nonexistent directory.
        """
        nonexistent_dir = tmp_path / "does_not_exist"

        async with mcp_client:
            result = await mcp_client.call_tool(
                "list_leak_reports", {"report_directory": str(nonexistent_dir)}
            )

        assert hasattr(result, "data"), f"Result {type(result)} has no data attribute"
        assert not result.is_error, "Tool call should not result in an error"
        assert len(result.data.summaries) == 0


class TestGetLeakReport:
    """
    Tests for the `get_leak_report` MCP tool.
    """

    @pytest.mark.asyncio
    async def test_get_report_basic(self, mcp_client: Client[Any], test_reports_dir: Path) -> None:
        """
        Test getting a basic report with default parameters.
        """
        async with mcp_client:
            result = await mcp_client.call_tool(
                "get_leak_report",
                {"report_id": "lyy0-uf3p", "report_directory": str(test_reports_dir)},
            )

        assert hasattr(result, "data"), f"Result {type(result)} has no data attribute"
        assert not result.is_error, "Tool call should not result in an error"
        report_data = result.data

        # Verify summary metadata
        assert report_data.summary.metadata.report_id == "lyy0-uf3p"
        assert report_data.summary.metadata.entrypoint == "utils_for_testing.py"
        assert report_data.summary.metadata.arguments == []
        assert report_data.summary.metadata.start_time == datetime(
            2025, 9, 29, 20, 54, 6, 560032, tzinfo=timezone.utc
        )
        assert report_data.summary.iteration_count == 3

        # Verify iterations (default num_iterations=1, so should get most recent iteration)
        assert len(report_data.iterations) == 1
        iteration = report_data.iterations[0]
        assert iteration.report_id == "lyy0-uf3p"
        assert iteration.iteration_number == 3

        # Verify memory usage
        assert iteration.memory_usage.current_rss_bytes == 356024320
        assert iteration.memory_usage.peak_rss_bytes == 376750080
        assert iteration.memory_usage.current_rss_diff == 73564160
        assert iteration.memory_usage.peak_rss_diff == 72286208
        assert iteration.memory_usage.diff_from_iteration == 2

        # Verify leak summary
        assert iteration.leak_summary.iteration == 3
        assert len(iteration.leak_summary.type_summaries) == 2
        assert iteration.leak_summary.type_summaries[0].object_type == "__main__.MemalotObject"
        assert iteration.leak_summary.type_summaries[0].count == 1
        assert iteration.leak_summary.type_summaries[1].object_type == "builtins.list"
        assert iteration.leak_summary.type_summaries[1].count == 1
        assert iteration.leak_summary.max_types_in_summary == 500

        # Verify object details
        assert len(iteration.object_details_list) == 2
        assert iteration.object_details_list[0].object_type_name == "__main__.MemalotObject"
        assert iteration.object_details_list[0].deep_size_bytes.approx_size == 1436
        assert iteration.object_details_list[0].deep_size_bytes.upper_bound_known is True
        assert iteration.object_details_list[0].referrers_checked is True
        assert iteration.object_details_list[1].object_type_name == "builtins.list"
        assert iteration.object_details_list[1].deep_size_bytes.approx_size == 1108
        assert iteration.object_details_list[1].deep_size_bytes.upper_bound_known is True
        assert iteration.object_details_list[1].referrers_checked is True

    @pytest.mark.asyncio
    async def test_get_report_multiple_iterations(
        self, mcp_client: Client[Any], test_reports_dir: Path
    ) -> None:
        """
        Test getting a report with multiple iterations.
        """
        async with mcp_client:
            result = await mcp_client.call_tool(
                "get_leak_report",
                {
                    "report_id": "lyy0-uf3p",
                    "num_iterations": 2,
                    "report_directory": str(test_reports_dir),
                },
            )

        assert hasattr(result, "data"), f"Result {type(result)} has no data attribute"
        assert not result.is_error, "Tool call should not result in an error"
        report_data = result.data

        # Verify summary (same as in test_get_report_basic)
        assert report_data.summary.metadata.report_id == "lyy0-uf3p"
        assert report_data.summary.iteration_count == 3

        # Verify we get 2 iterations (the most recent 2)
        assert len(report_data.iterations) == 2
        iteration_numbers = [iteration.iteration_number for iteration in report_data.iterations]
        assert iteration_numbers == [2, 3]  # Sorted by iteration number

        # Verify iteration 2
        iter_2 = report_data.iterations[0]
        assert iter_2.report_id == "lyy0-uf3p"
        assert iter_2.iteration_number == 2
        assert iter_2.memory_usage.current_rss_bytes == 282460160
        assert iter_2.memory_usage.peak_rss_bytes == 304463872
        assert iter_2.memory_usage.current_rss_diff == 215252992
        assert iter_2.memory_usage.peak_rss_diff == 235831296
        assert iter_2.memory_usage.diff_from_iteration == 1
        assert iter_2.leak_summary.iteration == 2
        assert len(iter_2.leak_summary.type_summaries) == 2
        assert iter_2.leak_summary.type_summaries[0].object_type == "__main__.MemalotObject"
        assert iter_2.leak_summary.type_summaries[0].count == 1
        assert len(iter_2.object_details_list) == 2

        # Verify iteration 3
        iter_3 = report_data.iterations[1]
        assert iter_3.report_id == "lyy0-uf3p"
        assert iter_3.iteration_number == 3
        assert iter_3.memory_usage.current_rss_bytes == 356024320
        assert iter_3.memory_usage.peak_rss_bytes == 376750080
        assert iter_3.leak_summary.iteration == 3
        assert len(iter_3.leak_summary.type_summaries) == 2
        assert len(iter_3.object_details_list) == 2

    @pytest.mark.asyncio
    async def test_get_report_with_filter_types(
        self, mcp_client: Client[Any], test_reports_dir: Path
    ) -> None:
        """
        Test getting a report with type filtering.
        """
        async with mcp_client:
            result = await mcp_client.call_tool(
                "get_leak_report",
                {
                    "report_id": "ykik-ab4l",
                    "report_directory": str(test_reports_dir),
                    "filter_types": "MemalotObject",
                },
            )

        assert hasattr(result, "data"), f"Result {type(result)} has no data attribute"
        assert not result.is_error, "Tool call should not result in an error"
        report_data = result.data

        # Verify summary metadata
        assert report_data.summary.metadata.report_id == "ykik-ab4l"
        assert report_data.summary.metadata.entrypoint == "utils_for_testing.py"
        assert report_data.summary.metadata.arguments == ["dummy_arg1", "dummy_arg2"]
        assert report_data.summary.iteration_count == 3

        # Verify we get 1 iteration (default num_iterations=1)
        assert len(report_data.iterations) == 1
        iteration = report_data.iterations[0]
        assert iteration.report_id == "ykik-ab4l"
        assert iteration.iteration_number == 3

        # Verify leak summary
        assert iteration.leak_summary.iteration == 3
        assert len(iteration.leak_summary.type_summaries) == 1
        assert iteration.leak_summary.type_summaries[0].object_type == "__main__.MemalotObject"
        assert iteration.leak_summary.type_summaries[0].count == 1

    @pytest.mark.asyncio
    async def test_get_report_invalid_id(
        self, mcp_client: Client[Any], test_reports_dir: Path
    ) -> None:
        """
        Test getting a report with an invalid/nonexistent ID.
        """
        async with mcp_client:
            with pytest.raises(ToolError, match="Report with ID does-not-exist not found"):
                _ = await mcp_client.call_tool(
                    "get_leak_report",
                    {"report_id": "does-not-exist", "report_directory": str(test_reports_dir)},
                )



import os
import resource
from sys import platform

import psutil
from pydantic.dataclasses import dataclass

from memalot.base import MemalotCount


@dataclass
class MemalotMemoryUsage:
    """
    Represents memory usage information at a particular point in time.
    """

    current_rss_bytes: int
    """
    The current resident set size (RSS) in bytes.
    """

    peak_rss_bytes: int | None
    """
    The peak resident set size (RSS) in bytes. This is `None` if not supported on the
    current platform.
    """

    system_percent_used: float
    """
    The percent of total physical system memory used.
    """

    iteration_number: int
    """
    The iteration number that this memory usage record was generated.
    """


def get_memory_usage(iteration: MemalotCount) -> MemalotMemoryUsage:
    """
    Gets memory usage information.
    """
    pid = os.getpid()
    p = psutil.Process(pid)
    memory_info = p.memory_info()
    if "linux" in platform:  # pragma: no cover
        # On Linux ru_maxrss is in kilobytes.
        peak_rss = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss * 1024
    elif "darwin" in platform:  # pragma: no cover
        # On Mac ru_maxrss is in bytes.
        peak_rss = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    elif "win32" in platform:  # pragma: no cover
        # On Windows, peak_wset should be the same as the max RSS on Linux and Mac.
        # This is based on https://psutil.readthedocs.io/en/latest/#psutil.Process.memory_info
        # and the Windows PROCESS_MEMORY_COUNTERS_EX structure. The docs for this
        # say it is "the peak working set size, in bytes".
        peak_rss = memory_info.peak_wset
    else:  # pragma: no cover
        peak_rss = None
    rss = memory_info.rss
    return MemalotMemoryUsage(
        current_rss_bytes=rss,
        peak_rss_bytes=peak_rss,
        system_percent_used=p.memory_percent(memtype="rss"),
        iteration_number=int(iteration),
    )



"""
Integration test for the cache example using time-based leak monitoring.
"""

from __future__ import annotations

import time
from pathlib import Path
from threading import Thread
from typing import Any

import numpy as np
import pytest

import memalot
from memalot.reports import FileReportReader
from tests_integration.utils import (
    assert_object_type_details,
    assert_referrer_graphs_have_root_names,
    assert_warmup_iteration,
    has_leak_iterations,
    read_single_report,
)


class Cache:
    """
    A cache that stores objects in a list.
    """

    def __init__(self) -> None:
        self.cache_list: list[Any] = []


class LeakyObject:
    """
    An object with a numpy array payload.
    """

    def __init__(self, size: int) -> None:
        self.payload = np.ones(size)


class NotLeakyObject:
    """
    An object that is not leaked.
    """

    def __init__(self, size: int) -> None:
        self.payload = np.ones(size)


def create_leaky_object() -> LeakyObject:
    """
    Creates a leaky object after creating a non-leaky object.
    """
    _ = NotLeakyObject(128)
    return LeakyObject(32)


def create_and_cache(cache: Cache) -> None:
    """
    Creates a leaky object and adds it to the cache.
    """
    cache.cache_list.append(create_leaky_object())


def create_non_leaky_object_and_cache(cache: Cache) -> None:
    """
    Creates a non-leaky object and adds it to the cache.
    """
    cache.cache_list.append(NotLeakyObject(32))


class LeakingThread(Thread):
    def __init__(self, iteration_time: float) -> None:
        super().__init__()
        self._stopped = False
        self._iteration_time = iteration_time

    def run(self) -> None:
        live_forever_cache = Cache()
        live_for_one_iteration_cache = Cache()
        while not self._stopped:
            create_and_cache(live_forever_cache)
            create_non_leaky_object_and_cache(live_for_one_iteration_cache)
            time.sleep(self._iteration_time)
            live_for_one_iteration_cache.cache_list.clear()

    def stop(self) -> None:
        self._stopped = True


@pytest.mark.integration
class TestCacheExampleTimeBased:
    """
    Tests for the cache example using time-based monitoring.
    """

    def test_cache_example_time_based(self, tmp_path: Path) -> None:
        """
        Tests that the cache example with time-based monitoring generates expected reports.

        This test verifies that:
        - start_leak_monitoring creates a background thread
        - Objects created after warmup are properly detected as leaks
        - Reports contain correct type summaries and object details
        - The monitoring thread can be stopped cleanly
        """
        expected_leak_iterations = 2

        reader = FileReportReader(report_directory=tmp_path)

        # Some objects live for only one iteration, so we need to set a higher max_object_lifetime
        # that the iteration time.
        iteration_time = 0.1
        max_object_lifetime = 1.0
        assert max_object_lifetime > iteration_time

        # when: continuously add leaky objects until we have at least one iteration with leaks
        stoppable = memalot.start_leak_monitoring(
            max_object_lifetime=max_object_lifetime,
            warmup_time=5.0,
            report_directory=tmp_path,
            max_object_details=4,
            str_max_length=200,
        )

        leaking_thread = LeakingThread(iteration_time=iteration_time)
        leaking_thread.start()

        # then: check for iterations with leaks
        try:
            deadline = time.monotonic() + 240.0
            while True:
                if has_leak_iterations(
                    reader=reader, expected_leak_iterations=expected_leak_iterations
                ):
                    break
                if time.monotonic() >= deadline:
                    raise TimeoutError("Timeout waiting for iterations with leaks.")
                time.sleep(iteration_time)
        finally:
            # then: stop the monitoring thread and leaking thread
            stoppable.stop()
            stoppable.join(timeout=60.0)
            leaking_thread.stop()
            leaking_thread.join(timeout=60.0)

        # Read and verify the report
        full_report = read_single_report(tmp_path=tmp_path)

        assert full_report.summary.iteration_count >= expected_leak_iterations, (
            f"Expected at least {expected_leak_iterations} iterations, "
            f"got {full_report.summary.iteration_count}"
        )

        # Verify the first iteration is warmup
        iteration_1 = full_report.iterations[0]
        assert iteration_1.iteration_number == 1
        assert_warmup_iteration(iteration=iteration_1)

        leak_iterations = [
            iteration
            for iteration in full_report.iterations
            if len(iteration.leak_summary.type_summaries) > 0
        ]
        assert len(leak_iterations) >= expected_leak_iterations

        for leak_iteration in leak_iterations[:expected_leak_iterations]:
            # Check type summaries - we expect LeakyObject and ndarray only
            type_summaries = leak_iteration.leak_summary.type_summaries
            actual_types = {summary.object_type for summary in type_summaries}

            # Verify that at minimum we have LeakyObject and ndarray
            required_types = {
                "tests_integration.test_time_based.LeakyObject",
                "numpy.ndarray",
            }
            assert actual_types == required_types

            object_details_list = list(leak_iteration.object_details_list)

            # Verify LeakyObject
            assert_object_type_details(
                object_details_list=object_details_list,
                object_type="tests_integration.test_time_based.LeakyObject",
                expected_count=2,
                expected_target_names={"LeakyObject (object)"},
            )

            # Verify ndarray
            assert_object_type_details(
                object_details_list=object_details_list,
                object_type="numpy.ndarray",
                expected_count=2,
                expected_target_names={"ndarray (object)"},
            )

            # Verify root names for LeakyObject and ndarray
            # Both should trace back to the cache
            expected_root_names = {
                "run.live_forever_cache (local)",
            }
            assert_referrer_graphs_have_root_names(
                object_details_list=object_details_list,
                expected_root_names=expected_root_names,
            )



from rich.color import Color
from rich.style import Style
from rich.theme import Theme

FUNCTION_NAME = "function_name"
MEMORY_ABSOLUTE = "memory_absolute"
MEMORY_INCREASE = "memory_increase"
MEMORY_DECREASE = "memory_decrease"
COUNT = "count"
OBJECT_DETAILS_TITLE = "object_graph_title"
SIZE_TITLE = "size_title"
REFERRERS_TITLE = "referrers_title"
STRING_REPR_TITLE = "string_repr_title"
REFERRER_NAME = "referrer_name"
REFERRER_SUFFIX_LEAF = "referrer_name_leaf"
REFERRER_SUFFIX_CYCLE = "referrer_name_cycle"
OBJECT_ID = "object_id"
TITLE = "title"
HEADING_BORDER = "heading_border"
TABLE_TITLE = "table_title"
TABLE_HEADER = "table_header"
TABLE_BORDER = "table_border"
END_OF_SECTION_TEXT = "end_of_section_text"
CLI_TABLE_TITLE = "cli_table_title"
CLI_TABLE_STYLE = "cli_table_style"
CLI_TABLE_HEADER = "cli_table_header"
CLI_LIST_REPORT_ID = "cli_list_report_id"
CLI_ITERATIONS_COLUMN = "cli_iterations_column"
CLI_NO_REPORTS_MESSAGE = "cli_no_reports_message"

DEFAULT_RICH_THEME = Theme(
    {
        FUNCTION_NAME: Style(color="blue"),
        MEMORY_ABSOLUTE: Style(color="cyan"),
        MEMORY_INCREASE: Style(color="red"),
        MEMORY_DECREASE: Style(color="green"),
        COUNT: Style(color="magenta"),
        OBJECT_DETAILS_TITLE: Style(),
        SIZE_TITLE: Style(color="yellow"),
        REFERRERS_TITLE: Style(color="yellow"),
        STRING_REPR_TITLE: Style(color="yellow"),
        REFERRER_NAME: Style(color=Color.default()),
        OBJECT_ID: Style(color="bright_black"),
        TITLE: Style(bold=True),
        HEADING_BORDER: Style(color="yellow"),
        TABLE_TITLE: Style(color="yellow"),
        TABLE_HEADER: Style(color=Color.default()),
        TABLE_BORDER: Style(color="yellow"),
        END_OF_SECTION_TEXT: Style(color="yellow"),
        REFERRER_SUFFIX_LEAF: Style(color="blue", italic=True),
        REFERRER_SUFFIX_CYCLE: Style(color="bright_black", italic=True),
        CLI_TABLE_TITLE: Style(color="yellow"),
        CLI_TABLE_STYLE: Style(color="yellow"),
        CLI_TABLE_HEADER: Style(color="yellow"),
        CLI_LIST_REPORT_ID: Style(color="cyan"),
        CLI_ITERATIONS_COLUMN: Style(color="magenta"),
        CLI_NO_REPORTS_MESSAGE: Style(color="yellow"),
    }
)
"""
The default theme for rich output.
"""



"""
General integration (functional) tests for the application.
"""

from __future__ import annotations

from typing import Any

import pytest

from memalot import leak_monitor


class LeakedObject:
    pass


output_chunks = []


class LeakyObject:
    def __init__(self) -> None:
        self._data: list[Any] = []

    @leak_monitor(
        output_func=lambda x: output_chunks.append(x),
        tee_console=True,
    )
    def add_data(self) -> None:
        item = LeakedObject()
        self._data.append(item)


@pytest.mark.integration
class TestOutput:
    """
    Integration tests that check that the correct output is produced when the application
    is run.

    Note: we don't do extensive testing of the exact output, as it's awkward to test.
    """

    def test_output_for_simple_example(self) -> None:
        """
        Tests that a simple example that leaks some memory periodically outputs the correct
        information. We can't test the exact output, but we can check the rough format.
        """
        # ruff: noqa: W291
        memalot_object = LeakyObject()

        for _ in range(2):
            memalot_object.add_data()

        expected_chunks = [
            "Memalot is performing leak monitoring for the add_data function",
            "Memalot Report (iteration 1) ",
            "This is a warmup iteration. Reports will be available from the next iteration.",
            "End of Memalot Report (iteration 1)",
            "Memalot Report (iteration 2)",
            """
             Possible New Leaks (iteration 2)              
╭─────────────────────────────────────────────────┬───────╮
│ Object Type                                     │ Count │
├─────────────────────────────────────────────────┼───────┤
│ tests_integration.test_integration.LeakedObject │     1 │
╰─────────────────────────────────────────────────┴───────╯
""",
            "Memalot is generating object details. This may take some time...",
            "Details for tests_integration.test_integration.LeakedObject",
            "End of Memalot Report (iteration 2)",
        ]

        for output_chunk, expected_chunk in zip(output_chunks, expected_chunks, strict=True):
            assert expected_chunk in output_chunk



#!/usr/bin/env python3
"""
CLI tool for listing memory reports.
"""

import argparse
import sys
from pathlib import Path
from typing import List

from rich import box
from rich.console import Console
from rich.table import Table

from memalot.options import Options
from memalot.output import get_output_writer
from memalot.reports import (
    FullReport,
    ReportIteration,
    ReportSummary,
    filter_iteration_by_types,
    get_report_reader,
)
from memalot.themes import (
    CLI_ITERATIONS_COLUMN,
    CLI_LIST_REPORT_ID,
    CLI_NO_REPORTS_MESSAGE,
    CLI_TABLE_HEADER,
    CLI_TABLE_STYLE,
    CLI_TABLE_TITLE,
    DEFAULT_RICH_THEME,
)


def _list_reports(num_reports: int = 5, report_directory: str | None = None) -> List[ReportSummary]:
    """
    List the most recent reports.

    Args:
        num_reports: The number of reports to return.
        report_directory: The directory to search for reports. If None, the
            default directory will be used.

    Returns:
        A list of ReportSummary objects.
    """
    reader = get_report_reader(
        report_directory=Path(report_directory) if report_directory else None
    )
    summaries = reader.get_report_summaries()
    summaries.sort(key=lambda s: s.metadata.start_time, reverse=True)
    return summaries[:num_reports]


def _print_report(
    report_id: str,
    num_iterations: int = 1,
    report_directory: str | None = None,
    force_terminal: bool | None = None,
    filter_types: List[str] | None = None,
    summary_only: bool = False,
) -> None:
    """
    Print a specific report to the console.

    Args:
        report_id: The ID of the report to print.
        num_iterations: The number of iterations to print. Defaults to 1 (last iteration).
        report_directory: The directory to search for reports. If None, the
            default directory will be used.
        force_terminal: Forces the use of terminal control codes for colors and formatting.
        filter_types: List of object type substrings to filter on. If provided, only
            objects whose type names contain any of these substrings will be included.
        summary_only: If True, only print the summary table and not the full iterations.
    """
    reader = get_report_reader(
        report_directory=Path(report_directory) if report_directory else None
    )
    full_report: FullReport = reader.get_full_report(
        report_id=report_id, num_iterations=num_iterations
    )

    options = Options(force_terminal=force_terminal)
    writer = get_output_writer(options)

    for iteration in full_report.iterations:
        if summary_only:
            iteration = ReportIteration(
                report_id=iteration.report_id,
                iteration_number=iteration.iteration_number,
                start_time=iteration.start_time,
                end_time=iteration.end_time,
                memory_usage=iteration.memory_usage,
                leak_summary=iteration.leak_summary,
                object_details_list=[],
            )
        if filter_types:
            filtered_iteration = filter_iteration_by_types(iteration, filter_types)
            writer.write(filtered_iteration)
        else:
            writer.write(iteration)


def _display_reports(summaries: List[ReportSummary], force_terminal: bool | None = None) -> None:
    """
    Display the report summaries using rich formatting.

    Args:
        summaries: List of report summaries to display.
        force_terminal: Forces the use of terminal control codes for colors and formatting.
    """
    console = Console(theme=DEFAULT_RICH_THEME, force_terminal=force_terminal)

    if not summaries:
        console.print("No reports found.", style=CLI_NO_REPORTS_MESSAGE)
        return

    table = Table(
        title="Reports",
        title_style=CLI_TABLE_TITLE,
        show_lines=True,
        style=CLI_TABLE_STYLE,
        header_style=CLI_TABLE_HEADER,
        box=box.ROUNDED,
    )

    table.add_column("Report ID", style=CLI_LIST_REPORT_ID, no_wrap=True)
    table.add_column("Iterations", style=CLI_ITERATIONS_COLUMN, justify="right", no_wrap=True)
    table.add_column("Start Time", no_wrap=True)
    table.add_column("Entrypoint", no_wrap=False, overflow="fold")
    table.add_column("Arguments", no_wrap=False, overflow="fold")

    for summary in summaries:
        # Format the start time nicely
        start_time_str = summary.metadata.start_time.strftime("%Y-%m-%d %H:%M:%S")

        # Format arguments - join them without truncation
        args_str = " ".join(summary.metadata.arguments)

        table.add_row(
            summary.metadata.report_id,
            str(summary.iteration_count),
            start_time_str,
            summary.metadata.entrypoint,
            args_str,
        )

    console.print(table)


def entrypoint(sys_args: List[str]) -> int:
    """
    Main CLI entry point.
    """
    parser = argparse.ArgumentParser(description="Memalot memory reports CLI", prog="memalot")

    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # List command
    list_parser = subparsers.add_parser("list", help="List recent reports")
    list_parser.add_argument(
        "--num-reports", type=int, default=10, help="Number of reports to display (default: 10)"
    )
    list_parser.add_argument(
        "--report-directory",
        type=str,
        help="Directory to search for reports (default: ~/.memalot/reports)",
    )
    list_parser.add_argument(
        "--force-terminal",
        action="store_true",
        help="Force the use of terminal control codes for colors and formatting",
    )

    # Print command
    print_parser = subparsers.add_parser("print", help="Print a specific report")
    print_parser.add_argument(
        "report_id",
        type=str,
        help="The ID of the report to print",
    )
    print_parser.add_argument(
        "--num-iterations",
        type=int,
        default=1,
        help="Number of iterations to print (default: 1, most recent)",
    )
    print_parser.add_argument(
        "--report-directory",
        type=str,
        help="Directory to search for reports (default: ~/.memalot/reports)",
    )
    print_parser.add_argument(
        "--force-terminal",
        action="store_true",
        help="Force the use of terminal control codes for colors and formatting",
    )
    print_parser.add_argument(
        "--filter-types",
        type=str,
        help="Filter by object types (comma-separated substrings, case-insensitive)",
    )
    print_parser.add_argument(
        "--summary-only",
        action="store_true",
        help="Only print the summary table and not object details",
    )

    args = parser.parse_args(sys_args)

    if args.command == "list":
        summaries = _list_reports(
            num_reports=args.num_reports, report_directory=args.report_directory
        )
        _display_reports(
            summaries, force_terminal=args.force_terminal if args.force_terminal else None
        )
    elif args.command == "print":
        # Parse comma-separated filter types
        filter_types = None
        if args.filter_types:
            filter_types = [t.strip() for t in args.filter_types.split(",") if t.strip()]

        _print_report(
            report_id=args.report_id,
            num_iterations=args.num_iterations,
            report_directory=args.report_directory,
            force_terminal=args.force_terminal if args.force_terminal else None,
            filter_types=filter_types,
            summary_only=args.summary_only,
        )
    else:  # pragma: no cover
        parser.print_help()
        return 1

    return 0


def main() -> None:  # pragma: no cover
    sys.exit(entrypoint(sys.argv[1:]))


if __name__ == "__main__":  # pragma: no cover
    main()



import gc
import inspect
from typing import Any, List, Set, Tuple

import networkx as nx
import numpy as np
import pytest
from referrers import ReferrerGraphNode

from memalot import utils
from memalot.base import ReferrerGraph, ReferrerNode
from memalot.utils import PrintableReferrerNode
from tests.utils_for_testing import create_mock, one


@pytest.fixture(name="custom_tag")
def _custom_tag_fixture() -> str:
    """
    Custom tag fixture for testing.
    """
    return "custom_tag"


@pytest.fixture(name="numpy_array_holder")
def _numpy_array_holder_fixture() -> "NumpyArrayHolder":
    """
    Fixture providing a NumpyArrayHolder instance.
    """
    return NumpyArrayHolder(128)


@pytest.fixture(name="tuple_holder")
def _tuple_holder_fixture() -> "TupleHolder":
    """
    Fixture providing a TupleHolder instance.
    """
    return TupleHolder(("string in tuple",))


class NumpyArrayHolder:
    def __init__(self, size: int) -> None:
        self.payload = np.ones(size)


class TupleHolder:
    def __init__(self, the_tuple: Tuple[Any, ...]) -> None:
        self.the_tuple = the_tuple


class TestAsMib:
    """
    Tests for the `as_mib` function.
    """

    @pytest.mark.parametrize(
        "num_bytes,expected",
        [
            (0, "[memory_absolute]0.0[/memory_absolute] MiB"),
            (1024, "[memory_absolute]0.0[/memory_absolute] MiB"),
            (1048576, "[memory_absolute]1.0[/memory_absolute] MiB"),  # 1 MiB
            (2097152, "[memory_absolute]2.0[/memory_absolute] MiB"),  # 2 MiB
            (1572864, "[memory_absolute]1.5[/memory_absolute] MiB"),  # 1.5 MiB
            (1073741824, "[memory_absolute]1024.0[/memory_absolute] MiB"),  # 1 GiB
        ],
    )
    def test_as_mib_default_tag(self, num_bytes: int, expected: str) -> None:
        """
        Test as_mib with default tag and various byte values.
        """
        result = utils.as_mib(num_bytes)
        assert result == expected

    def test_as_mib_custom_tag(self, custom_tag: str) -> None:
        """
        Test as_mib with custom tag.
        """
        result = utils.as_mib(1048576, custom_tag)
        expected = f"[{custom_tag}]1.0[/{custom_tag}] MiB"
        assert result == expected


class TestAsMibSf:
    """
    Tests for the `as_mib_sf` function.
    """

    @pytest.mark.parametrize(
        "num_bytes,expected",
        [
            (0, "[memory_absolute]0[/memory_absolute] MiB"),
            (1024, "[memory_absolute]0.000977[/memory_absolute] MiB"),
            (1048576, "[memory_absolute]1[/memory_absolute] MiB"),  # 1 MiB
            (2097152, "[memory_absolute]2[/memory_absolute] MiB"),  # 2 MiB
            (1572864, "[memory_absolute]1.5[/memory_absolute] MiB"),  # 1.5 MiB
            (1073741824, "[memory_absolute]1.02e+03[/memory_absolute] MiB"),  # 1 GiB
        ],
    )
    def test_as_mib_sf_default_tag(self, num_bytes: int, expected: str) -> None:
        """
        Test as_mib_sf with default tag and various byte values.
        """
        result = utils.as_mib_sf(num_bytes)
        assert result == expected

    def test_as_mib_sf_custom_tag(self, custom_tag: str) -> None:
        """
        Test as_mib_sf with custom tag.
        """
        result = utils.as_mib_sf(1048576, custom_tag)
        expected = f"[{custom_tag}]1[/{custom_tag}] MiB"
        assert result == expected


class TestFormatBytes:
    """
    Tests for the `format_bytes` function.
    """

    @pytest.mark.parametrize(
        "num_bytes,expected",
        [
            (0, "0 B"),
            (512, "[memory_absolute]512[/memory_absolute] B"),
            (1023, "[memory_absolute]1023[/memory_absolute] B"),
            (1024, "[memory_absolute]1.0[/memory_absolute] KiB"),
            (1536, "[memory_absolute]1.5[/memory_absolute] KiB"),
            (1048576, "[memory_absolute]1.00[/memory_absolute] MiB"),
            (1572864, "[memory_absolute]1.50[/memory_absolute] MiB"),
            (1073741824, "[memory_absolute]1.00[/memory_absolute] GiB"),
            (1610612736, "[memory_absolute]1.50[/memory_absolute] GiB"),
        ],
    )
    def test_format_bytes_default_tag(self, num_bytes: int, expected: str) -> None:
        """
        Test format_bytes with default tag and various byte values.
        """
        result = utils.format_bytes(num_bytes)
        assert result == expected

    def test_format_bytes_custom_tag(self, custom_tag: str) -> None:
        """
        Test format_bytes with custom tag.
        """
        result = utils.format_bytes(1024, custom_tag)
        expected = f"[{custom_tag}]1.0[/{custom_tag}] KiB"
        assert result == expected

    @pytest.mark.parametrize(
        "invalid_input",
        [
            -1,
            -100,
            "1024",
            1024.5,
            None,
        ],
    )
    def test_format_bytes_invalid_input(self, invalid_input: Any) -> None:
        """
        Test format_bytes raises ValueError for invalid inputs.
        """
        with pytest.raises(ValueError, match="Input must be a non-negative integer"):
            utils.format_bytes(invalid_input)


class TestGetFullTypeName:
    """
    Tests for the `get_full_type_name` function.
    """

    def test_get_full_type_name_builtin_types(self) -> None:
        """
        Test get_full_type_name with built-in types.
        """
        assert utils.get_full_type_name(str) == "builtins.str"
        assert utils.get_full_type_name(int) == "builtins.int"
        assert utils.get_full_type_name(list) == "builtins.list"
        assert utils.get_full_type_name(dict) == "builtins.dict"

    def test_get_full_type_name_custom_classes(self) -> None:
        """
        Test get_full_type_name with custom classes.
        """
        assert utils.get_full_type_name(NumpyArrayHolder) == "tests.test_utils.NumpyArrayHolder"
        assert utils.get_full_type_name(TupleHolder) == "tests.test_utils.TupleHolder"

    def test_get_full_type_name_imported_types(self) -> None:
        """
        Test get_full_type_name with imported types.
        """
        assert utils.get_full_type_name(np.ndarray) == "numpy.ndarray"
        assert utils.get_full_type_name(nx.DiGraph) == "networkx.classes.digraph.DiGraph"


class TestGetModulePrefix:
    """
    Tests for the `get_module_prefix` function.
    """

    def test_get_module_prefix_found(self) -> None:
        """
        Test get_module_prefix returns the correct module prefix when called from a module
        outside the memalot package.
        """
        # When this test runs, the calling module is tests.test_utils, so we expect "tests."
        result = utils.get_module_prefix()
        assert result == "tests."


class _ObjectThatOnlyHasOneInstance:
    pass


class TestGetObjects:
    """
    Tests for the `get_objects` function.
    """

    def test_get_objects_no_mocking(self) -> None:
        """
        Test get_objects works without mocking get_objects_func. We don't know what the result
        will be in this case.
        """
        _ = _ObjectThatOnlyHasOneInstance()
        result = utils.get_objects()
        instances = [obj for obj in result if isinstance(obj, _ObjectThatOnlyHasOneInstance)]
        assert len(instances) == 1

    def test_get_objects_mocked_empty_list(self) -> None:
        """
        Test get_objects with mocked get_objects_func returning empty list.
        """
        mock_get_objects = create_mock(spec=lambda: [])
        mock_get_objects.return_value = []

        result = utils.get_objects(get_objects_func=mock_get_objects)

        assert result == []
        mock_get_objects.assert_called_once()

    def test_get_objects_mocked_with_objects(self) -> None:
        """
        Test get_objects with mocked get_objects_func returning known objects.
        """
        test_string = "test_string"
        test_list = [1, 2, 3]
        test_dict = {"key": "value"}

        mock_get_objects = create_mock(spec=lambda: [])
        mock_get_objects.return_value = [test_string, test_list, test_dict, 1, 2, 3, "key", "value"]

        result = utils.get_objects(get_objects_func=mock_get_objects)

        assert len(result) == 8
        assert test_string in result
        assert test_list in result
        assert test_dict in result
        assert "key" in result
        assert "value" in result
        assert 1 in result
        assert 2 in result
        assert 3 in result
        mock_get_objects.assert_called_once()

    def test_get_objects_mocked_with_frame_object(self) -> None:
        """
        Test get_objects filters out frame objects when using mocked get_objects_func.
        """
        test_string = "test_string"
        frame = inspect.currentframe()

        mock_get_objects = create_mock(spec=lambda: [])
        mock_get_objects.return_value = [test_string, frame]

        result = utils.get_objects(get_objects_func=mock_get_objects)

        assert len(result) == 1
        assert test_string in result
        assert frame not in result
        mock_get_objects.assert_called_once()

    def test_get_objects_mocked_with_untracked_objects(self) -> None:
        """
        Test get_objects finds untracked objects through referents when using mocked
        get_objects_func.
        """
        # Create a tracked object (list) that contains untracked objects (tuple)
        untracked_tuple = ("untracked", "data")
        tracked_list = [untracked_tuple]

        mock_get_objects = create_mock(spec=lambda: [])
        mock_get_objects.return_value = [tracked_list]

        result = utils.get_objects(max_untracked_search_depth=2, get_objects_func=mock_get_objects)

        result_ids = {id(obj) for obj in result}
        assert id(tracked_list) in result_ids
        assert id(untracked_tuple) in result_ids
        mock_get_objects.assert_called_once()

    def test_get_objects_removes_duplicates(self) -> None:
        """
        Test get_objects removes duplicate objects when using mocked get_objects_func.
        """
        test_string = "test_string"

        mock_get_objects = create_mock(spec=lambda: [])
        mock_get_objects.return_value = [test_string, test_string, test_string]

        result = utils.get_objects(get_objects_func=mock_get_objects)

        # Should only have one instance of the string
        string_count = sum(1 for obj in result if obj is test_string)
        assert string_count == 1
        mock_get_objects.assert_called_once()

    def test_get_objects_with_np_array_holder(self, numpy_array_holder: NumpyArrayHolder) -> None:
        """
        Tests that NumPy arrays are returned when they are instance attributes of a class.

        In Python 3.10, for example, instance attributes are held in a dict, and since CPython
        doesn't track collections of immutable objects, we need to search the referents for
        at least two levels to find the NumPy array in this case.
        """
        objects = utils.get_objects()
        object_ids: Set[int] = {id(obj) for obj in objects}
        assert id(numpy_array_holder.payload) in object_ids

    def test_get_objects_with_tuple_holder(self, tuple_holder: TupleHolder) -> None:
        """
        Tests that tuples are returned when they are instance attributes of a class.

        In Python 3.10, for example, instance attributes are held in a dict, and since CPython
        doesn't track collections of immutable objects, we need to search the referents for
        at least two levels to find the tuple in this case.
        """
        objects = utils.get_objects()
        object_ids: Set[int] = {id(obj) for obj in objects}
        assert id(tuple_holder.the_tuple) in object_ids
        assert id(tuple_holder.the_tuple[0]) in object_ids

    def test_with_nested_tuples(self) -> None:
        """
        Tests the `get_objects` function with nested tuples. Nested tuples are interesting
        because they are both immutable and collections. Since CPython doesn't track
        collections of immutable objects, it's possible for nested tuples to "hide" objects
        from `get_objects`.

        Note: we need to increase `max_untracked_search_depth` to 5 to find the innermost
        tuple. This is because (in Python 3.10) we have the following reference chain:

        TupleHolder -> __dict__ -> Tuple c -> Tuple b -> Tuple a -> string
        """
        my_object = TupleHolder(self._get_nested_tuples())
        objects = utils.get_objects(max_untracked_search_depth=5)
        object_ids: Set[int] = {id(obj) for obj in objects}
        assert id(my_object.the_tuple) in object_ids
        assert id(my_object.the_tuple[0]) in object_ids
        assert id(my_object.the_tuple[0][0]) in object_ids
        assert id(my_object.the_tuple[0][0][0]) in object_ids

    def _get_nested_tuples(self) -> Tuple[Tuple[Tuple[str]]]:
        a = ("tuples all the way down",)
        b = (a,)
        c = (b,)
        # CPython stops tracking tuples if they contain only immutable objects, but
        # only when they are first seen by the garbage collector, so we need to collect
        # here to trigger this.
        gc.collect()
        return c


class TestConvertGraphNodes:
    """
    Tests for the `convert_graph_nodes` function.
    """

    def test_convert_graph_nodes_empty_graph(self) -> None:
        """
        Test convert_graph_nodes with empty graph.
        """
        empty_graph = nx.DiGraph()
        result = utils.convert_graph_nodes(empty_graph)

        assert isinstance(result, ReferrerGraph)
        assert len(result.graph_nodes) == 0
        assert result.target_ids == set()
        assert result.root_ids == set()

    @pytest.mark.parametrize(
        "node,expected",
        [
            (
                ReferrerGraphNode(name="test_node", id=123, type="str"),
                ReferrerNode(
                    id=0,
                    name="test_node",
                    object_id=123,
                    referent_ids=[],
                ),
            ),
        ],
    )
    def test_convert_graph_nodes_single_node(
        self, node: ReferrerGraphNode, expected: ReferrerNode
    ) -> None:
        """
        Test convert_graph_nodes with single node.
        """
        graph = nx.DiGraph()
        graph.add_node(node)

        result = utils.convert_graph_nodes(graph)

        assert len(result.graph_nodes) == 1
        assert result.graph_nodes[0] == expected
        # Single node with no edges should be a root but not a target
        assert result.root_ids == {0}
        assert result.target_ids == set()

    def test_convert_graph_nodes_with_edges(self) -> None:
        """
        Test convert_graph_nodes with nodes and edges.
        """
        graph = nx.DiGraph()
        node1 = ReferrerGraphNode(name="test_node1", id=123, type="str")
        node2 = ReferrerGraphNode(name="test_node2", id=456, type="str")
        graph.add_node(node1)
        graph.add_node(node2)
        graph.add_edge(node1, node2)

        result = utils.convert_graph_nodes(graph)

        assert len(result.graph_nodes) == 2

        # Check that the first node has the second node as a referent
        assert result.graph_nodes[0].referent_ids == [1]
        assert result.graph_nodes[1].referent_ids == []

        # Verify node details
        assert result.graph_nodes[0].name == "test_node1"
        assert result.graph_nodes[0].object_id == 123
        assert result.graph_nodes[1].name == "test_node2"
        assert result.graph_nodes[1].object_id == 456

        # Node1 should be root (no incoming edges), node2 should not be root
        assert result.root_ids == {0}
        assert result.target_ids == set()

    def test_convert_graph_nodes_with_target_nodes(self) -> None:
        """
        Test convert_graph_nodes correctly identifies target nodes.
        """
        graph = nx.DiGraph()
        node1 = ReferrerGraphNode(name="root_node", id=123, type="str")
        node2 = ReferrerGraphNode(name="target_node", id=456, type="str")
        # Manually set is_target attribute for testing
        object.__setattr__(node2, "is_target", True)
        graph.add_node(node1)
        graph.add_node(node2)
        graph.add_edge(node1, node2)

        result = utils.convert_graph_nodes(graph)

        assert len(result.graph_nodes) == 2

        # Verify target and root identification
        assert result.root_ids == {0}  # node1 is root (no incoming edges)
        assert result.target_ids == {1}  # node2 is target (is_target=True)

        # Verify node details
        assert result.graph_nodes[0].name == "root_node"
        assert result.graph_nodes[1].name == "target_node"

    def test_convert_graph_nodes_with_multiple_targets_and_roots(self) -> None:
        """
        Test convert_graph_nodes with multiple target and root nodes.
        """
        graph = nx.DiGraph()
        root1 = ReferrerGraphNode(name="root1", id=100, type="str")
        root2 = ReferrerGraphNode(name="root2", id=200, type="str")
        target1 = ReferrerGraphNode(name="target1", id=300, type="str")
        target2 = ReferrerGraphNode(name="target2", id=400, type="str")

        # Manually set is_target attribute for testing
        object.__setattr__(target1, "is_target", True)
        object.__setattr__(target2, "is_target", True)

        graph.add_node(root1)
        graph.add_node(root2)
        graph.add_node(target1)
        graph.add_node(target2)
        graph.add_edge(root1, target1)
        graph.add_edge(root2, target2)

        result = utils.convert_graph_nodes(graph)

        assert len(result.graph_nodes) == 4

        # Verify multiple roots and targets
        assert set(result.root_ids) == {0, 1}  # root1 and root2
        assert set(result.target_ids) == {2, 3}  # target1 and target2

    def test_convert_graph_nodes_with_replacements(self) -> None:
        """
        Test convert_graph_nodes applies replacements to node names.
        """
        graph = nx.DiGraph()
        node1 = ReferrerGraphNode(name="module.Class.method", id=123, type="str")
        node2 = ReferrerGraphNode(name="module.AnotherClass", id=456, type="str")
        graph.add_node(node1)
        graph.add_node(node2)
        graph.add_edge(node1, node2)

        replacements = {"module.": "pkg.", "Class": "Klass"}
        result = utils.convert_graph_nodes(graph, replacements)

        assert len(result.graph_nodes) == 2
        assert result.graph_nodes[0].name == "pkg.Klass.method"
        assert result.graph_nodes[1].name == "pkg.AnotherKlass"

    def test_convert_graph_nodes_invalid_node_type(self) -> None:
        """
        Test convert_graph_nodes raises ValueError for invalid node types.
        """
        graph = nx.DiGraph()
        graph.add_node("invalid_node")

        with pytest.raises(ValueError, match="Unexpected type"):
            utils.convert_graph_nodes(graph)

    def test_convert_graph_nodes_invalid_edge_type(self) -> None:
        """
        Test convert_graph_nodes raises ValueError for invalid edge types.
        """
        graph = nx.DiGraph()
        node1 = ReferrerGraphNode(name="test_node", id=123, type="str")
        graph.add_node(node1)
        graph.add_node("invalid_node")
        graph.add_edge(node1, "invalid_node")

        with pytest.raises(ValueError, match="Unexpected type"):
            utils.convert_graph_nodes(graph)


class TestPluralize:
    """
    Tests for the `pluralize` function.
    """

    @pytest.mark.parametrize(
        "base_string,expected",
        [
            pytest.param("item", "item", id="simple_word"),
            pytest.param("", "", id="empty_string"),
        ],
    )
    def test_pluralize_singular_form(self, base_string: str, expected: str) -> None:
        """
        Test pluralize returns the base string unchanged when num is 1.
        """
        result = utils.pluralize(base_string, 1)
        assert result == expected

    @pytest.mark.parametrize(
        "base_string,num,expected",
        [
            pytest.param("item", 0, "items", id="zero_items"),
            pytest.param("item", 2, "items", id="two_items"),
            pytest.param("item", 100, "items", id="many_items"),
        ],
    )
    def test_pluralize_plural_form(self, base_string: str, num: int, expected: str) -> None:
        """
        Test pluralize adds 's' to the base string when num is not 1.
        """
        result = utils.pluralize(base_string, num)
        assert result == expected


class TestConvertGraphNodesToPrintable:
    """
    Tests for the `convert_graph_nodes_to_printable` function.
    """

    def test_graph_with_multiple_targets(self) -> None:
        """
        Test convert_graph_nodes_to_printable with a graph containing multiple target nodes.

        Graph shape: Two separate chains, each with a target node.
        root1 -> target1
        root2 -> target2
        """
        # given
        nodes = [
            ReferrerNode(id=0, name="root1", object_id=100, referent_ids=[1]),
            ReferrerNode(id=1, name="target1", object_id=200, referent_ids=[]),
            ReferrerNode(id=2, name="root2", object_id=300, referent_ids=[3]),
            ReferrerNode(id=3, name="target2", object_id=400, referent_ids=[]),
        ]
        referrer_graph = ReferrerGraph(target_ids={1, 3}, root_ids={0, 2}, graph_nodes=nodes)

        # when
        result = utils.convert_graph_nodes_to_printable(referrer_graph)

        # then
        result_nodes = list(result.nodes())
        assert len(result_nodes) == 4

        # Verify node type counts
        self._verify_node_counts_by_type(
            nodes=result_nodes,
            expected_targets=2,
            expected_roots=2,
            expected_cycle_members=0,
            expected_regular=0,
        )

        # Verify each node's properties individually
        root1_nodes = self._find_nodes_by_name(result_nodes, "root1")
        root1_node = one(root1_nodes)
        self._assert_node_properties(
            node=root1_node,
            expected_name="root1",
            expected_object_id=100,
            expected_is_cycle_member=False,
            expected_is_root=True,
            expected_is_target=False,
        )

        target1_nodes = self._find_nodes_by_name(result_nodes, "target1")
        target1_node = one(target1_nodes)
        self._assert_node_properties(
            node=target1_node,
            expected_name="target1",
            expected_object_id=200,
            expected_is_cycle_member=False,
            expected_is_root=False,
            expected_is_target=True,
        )

        root2_nodes = self._find_nodes_by_name(result_nodes, "root2")
        root2_node = one(root2_nodes)
        self._assert_node_properties(
            node=root2_node,
            expected_name="root2",
            expected_object_id=300,
            expected_is_cycle_member=False,
            expected_is_root=True,
            expected_is_target=False,
        )

        target2_nodes = self._find_nodes_by_name(result_nodes, "target2")
        target2_node = one(target2_nodes)
        self._assert_node_properties(
            node=target2_node,
            expected_name="target2",
            expected_object_id=400,
            expected_is_cycle_member=False,
            expected_is_root=False,
            expected_is_target=True,
        )

        # Verify graph edges (remember: edges are reversed in the printable graph)
        # Original: root1 -> target1, root2 -> target2
        # Reversed: target1 -> root1, target2 -> root2
        self._verify_node_successors(result, target1_node, ["root1"])
        self._verify_node_successors(result, target2_node, ["root2"])
        self._verify_node_successors(result, root1_node, [])  # Roots have no successors
        self._verify_node_successors(result, root2_node, [])  # Roots have no successors

        # Verify predecessors (opposite direction)
        self._verify_node_predecessors(result, root1_node, ["target1"])
        self._verify_node_predecessors(result, root2_node, ["target2"])
        self._verify_node_predecessors(result, target1_node, [])  # Targets have no predecessors
        self._verify_node_predecessors(result, target2_node, [])  # Targets have no predecessors

    def test_graph_with_multiple_roots(self) -> None:
        """
        Test convert_graph_nodes_to_printable with a graph containing multiple root nodes.

        Graph shape: pattern with two roots converging to one target.
        root1 -> intermediate1 -> target
        root2 -> intermediate2 -> target
        """
        # given
        nodes = [
            ReferrerNode(id=0, name="root1", object_id=100, referent_ids=[2]),
            ReferrerNode(id=1, name="root2", object_id=200, referent_ids=[3]),
            ReferrerNode(id=2, name="intermediate1", object_id=300, referent_ids=[4]),
            ReferrerNode(id=3, name="intermediate2", object_id=400, referent_ids=[4]),
            ReferrerNode(id=4, name="target", object_id=500, referent_ids=[]),
        ]
        referrer_graph = ReferrerGraph(target_ids={4}, root_ids={0, 1}, graph_nodes=nodes)

        # when
        result = utils.convert_graph_nodes_to_printable(referrer_graph)

        # then
        result_nodes = list(result.nodes())

        # Should have 5 nodes: no duplicates created since no single node has multiple referent_ids
        assert len(result_nodes) == 5

        # Verify node type counts
        self._verify_node_counts_by_type(
            nodes=result_nodes,
            expected_targets=1,
            expected_roots=2,
            # No cycle members since no single node has multiple referent_ids
            expected_cycle_members=0,
            expected_regular=2,  # Two intermediate nodes
        )

        # Verify root nodes
        root1_nodes = self._find_nodes_by_name(result_nodes, "root1")
        root1_node = one(root1_nodes)
        self._assert_node_properties(
            node=root1_node,
            expected_name="root1",
            expected_object_id=100,
            expected_is_cycle_member=False,
            expected_is_root=True,
            expected_is_target=False,
        )

        root2_nodes = self._find_nodes_by_name(result_nodes, "root2")
        root2_node = one(root2_nodes)
        self._assert_node_properties(
            node=root2_node,
            expected_name="root2",
            expected_object_id=200,
            expected_is_cycle_member=False,
            expected_is_root=True,
            expected_is_target=False,
        )

        # Verify target node
        target_nodes = self._find_nodes_by_name(result_nodes, "target")
        target_node = one(target_nodes)
        self._assert_node_properties(
            node=target_node,
            expected_name="target",
            expected_object_id=500,
            expected_is_cycle_member=False,
            expected_is_root=False,
            expected_is_target=True,
        )

        # Verify intermediate nodes (should have one of each, no duplicates)
        intermediate1_nodes = self._find_nodes_by_name(result_nodes, "intermediate1")
        intermediate1_node = one(intermediate1_nodes)
        self._assert_node_properties(
            node=intermediate1_node,
            expected_name="intermediate1",
            expected_object_id=300,
            expected_is_cycle_member=False,
            expected_is_root=False,
            expected_is_target=False,
        )

        intermediate2_nodes = self._find_nodes_by_name(result_nodes, "intermediate2")
        intermediate2_node = one(intermediate2_nodes)
        self._assert_node_properties(
            node=intermediate2_node,
            expected_name="intermediate2",
            expected_object_id=400,
            expected_is_cycle_member=False,
            expected_is_root=False,
            expected_is_target=False,
        )

        # Verify graph edges (remember: edges are reversed in the printable graph)
        # Original: root1 -> intermediate1 -> target, root2 -> intermediate2 -> target
        # Reversed: target -> intermediate1 -> root1, target -> intermediate2 -> root2
        self._verify_node_successors(result, target_node, ["intermediate1", "intermediate2"])
        self._verify_node_successors(result, intermediate1_node, ["root1"])
        self._verify_node_successors(result, intermediate2_node, ["root2"])
        self._verify_node_successors(result, root1_node, [])  # Roots have no successors
        self._verify_node_successors(result, root2_node, [])  # Roots have no successors

        # Verify predecessors
        self._verify_node_predecessors(result, root1_node, ["intermediate1"])
        self._verify_node_predecessors(result, root2_node, ["intermediate2"])
        self._verify_node_predecessors(result, intermediate1_node, ["target"])
        self._verify_node_predecessors(result, intermediate2_node, ["target"])
        self._verify_node_predecessors(result, target_node, [])  # Target has no predecessors

    def test_graph_with_multiple_referrers_to_nodes(self) -> None:
        """
        Test convert_graph_nodes_to_printable with nodes that have multiple referrers.

        Graph shape: Multiple nodes referring to the same intermediate node.
        root1 -> shared_node -> target
        root2 -> shared_node (same object as above)

        This test verifies that the function handles multiple nodes referring to the same object
        without creating cycle members (since no single node has multiple referent_ids).
        """
        # given
        nodes = [
            ReferrerNode(id=0, name="root1", object_id=100, referent_ids=[2]),
            ReferrerNode(id=1, name="root2", object_id=200, referent_ids=[2]),
            ReferrerNode(id=2, name="shared_node", object_id=300, referent_ids=[3]),
            ReferrerNode(id=3, name="target", object_id=400, referent_ids=[]),
        ]
        referrer_graph = ReferrerGraph(target_ids={3}, root_ids={0, 1}, graph_nodes=nodes)

        # when
        result = utils.convert_graph_nodes_to_printable(referrer_graph)

        # then
        result_nodes = list(result.nodes())

        # Should have exactly 4 nodes since no single node has multiple referent_ids
        assert len(result_nodes) == 4

        # Verify node type counts
        self._verify_node_counts_by_type(
            nodes=result_nodes,
            expected_targets=1,
            expected_roots=2,
            expected_cycle_members=0,  # No cycle members expected
            expected_regular=1,  # One shared node
        )

        # Verify each node's properties individually
        root1_nodes = self._find_nodes_by_name(result_nodes, "root1")
        root1_node = one(root1_nodes)
        self._assert_node_properties(
            node=root1_node,
            expected_name="root1",
            expected_object_id=100,
            expected_is_cycle_member=False,
            expected_is_root=True,
            expected_is_target=False,
        )

        root2_nodes = self._find_nodes_by_name(result_nodes, "root2")
        root2_node = one(root2_nodes)
        self._assert_node_properties(
            node=root2_node,
            expected_name="root2",
            expected_object_id=200,
            expected_is_cycle_member=False,
            expected_is_root=True,
            expected_is_target=False,
        )

        shared_node_nodes = self._find_nodes_by_name(result_nodes, "shared_node")
        shared_node = one(shared_node_nodes)
        self._assert_node_properties(
            node=shared_node,
            expected_name="shared_node",
            expected_object_id=300,
            expected_is_cycle_member=False,
            expected_is_root=False,
            expected_is_target=False,
        )

        target_nodes = self._find_nodes_by_name(result_nodes, "target")
        target_node = one(target_nodes)
        self._assert_node_properties(
            node=target_node,
            expected_name="target",
            expected_object_id=400,
            expected_is_cycle_member=False,
            expected_is_root=False,
            expected_is_target=True,
        )

        # Verify graph edges (remember: edges are reversed in the printable graph)
        # Original: root1 -> shared_node -> target, root2 -> shared_node -> target
        # Reversed: target -> shared_node -> root1, target -> shared_node -> root2
        self._verify_node_successors(result, target_node, ["shared_node"])
        self._verify_node_successors(result, shared_node, ["root1", "root2"])
        self._verify_node_successors(result, root1_node, [])  # Roots have no successors
        self._verify_node_successors(result, root2_node, [])  # Roots have no successors

        # Verify predecessors
        self._verify_node_predecessors(result, root1_node, ["shared_node"])
        self._verify_node_predecessors(result, root2_node, ["shared_node"])
        self._verify_node_predecessors(result, shared_node, ["target"])
        self._verify_node_predecessors(result, target_node, [])  # Target has no predecessors

    def test_cycle_detection_and_breaking(self) -> None:
        """
        Test convert_graph_nodes_to_printable handles multiple referent_ids by creating
        duplicate nodes.

        Graph shape: Two nodes with multiple referent_ids create cycle members.
        root -> node1, node2 (root has 2 referents)
        node1 -> target1, target2, intermediate (node1 has 3 referents)
        node2 -> intermediate
        """
        # given
        nodes = [
            ReferrerNode(id=0, name="root", object_id=100, referent_ids=[1, 2]),  # 2 referents
            ReferrerNode(id=1, name="node1", object_id=200, referent_ids=[3, 4, 5]),  # 3 referents
            ReferrerNode(id=2, name="node2", object_id=250, referent_ids=[5]),
            ReferrerNode(id=3, name="target1", object_id=300, referent_ids=[]),
            ReferrerNode(id=4, name="target2", object_id=400, referent_ids=[]),
            ReferrerNode(id=5, name="intermediate", object_id=500, referent_ids=[]),
        ]
        referrer_graph = ReferrerGraph(target_ids={3, 4}, root_ids={0}, graph_nodes=nodes)

        # when
        result = utils.convert_graph_nodes_to_printable(referrer_graph)

        # then
        result_nodes = list(result.nodes())

        # Should have 9 nodes: original 6 + 3 duplicates (1 root + 2xnode1)
        assert len(result_nodes) == 9

        # Verify node type counts
        self._verify_node_counts_by_type(
            nodes=result_nodes,
            expected_targets=2,  # target1, target2
            expected_roots=2,  # Original root + duplicate root (cycle member)
            expected_cycle_members=3,  # 1 duplicate root + 2 duplicate node1s
            expected_regular=3,  # node1, node2, intermediate (originals only)
        )

        # Verify root nodes (should have original + duplicate)
        root_nodes = self._find_nodes_by_name(result_nodes, "root")
        assert len(root_nodes) == 2

        # Find original and duplicate root
        original_root = self._get_node_by_name_and_cycle_status(result_nodes, "root", False)
        duplicate_root = self._get_node_by_name_and_cycle_status(result_nodes, "root", True)

        self._assert_node_properties(
            node=original_root,
            expected_name="root",
            expected_object_id=100,
            expected_is_cycle_member=False,
            expected_is_root=True,
            expected_is_target=False,
        )

        self._assert_node_properties(
            node=duplicate_root,
            expected_name="root",
            expected_object_id=100,
            expected_is_cycle_member=True,
            expected_is_root=True,
            expected_is_target=False,
        )

        # Verify node1 nodes (should have original + 2 duplicates)
        node1_nodes = self._find_nodes_by_name(result_nodes, "node1")
        assert len(node1_nodes) == 3

        original_node1 = self._get_node_by_name_and_cycle_status(result_nodes, "node1", False)
        duplicate_node1s = [node for node in node1_nodes if node.is_cycle_member]
        assert len(duplicate_node1s) == 2

        self._assert_node_properties(
            node=original_node1,
            expected_name="node1",
            expected_object_id=200,
            expected_is_cycle_member=False,
            expected_is_root=False,
            expected_is_target=False,
        )

        for duplicate_node1 in duplicate_node1s:
            self._assert_node_properties(
                node=duplicate_node1,
                expected_name="node1",
                expected_object_id=200,
                expected_is_cycle_member=True,
                expected_is_root=False,
                expected_is_target=False,
            )

        # Verify other nodes
        node2_nodes = self._find_nodes_by_name(result_nodes, "node2")
        node2_node = one(node2_nodes)
        self._assert_node_properties(
            node=node2_node,
            expected_name="node2",
            expected_object_id=250,
            expected_is_cycle_member=False,
            expected_is_root=False,
            expected_is_target=False,
        )

        target1_nodes = self._find_nodes_by_name(result_nodes, "target1")
        target1_node = one(target1_nodes)
        self._assert_node_properties(
            node=target1_node,
            expected_name="target1",
            expected_object_id=300,
            expected_is_cycle_member=False,
            expected_is_root=False,
            expected_is_target=True,
        )

        target2_nodes = self._find_nodes_by_name(result_nodes, "target2")
        target2_node = one(target2_nodes)
        self._assert_node_properties(
            node=target2_node,
            expected_name="target2",
            expected_object_id=400,
            expected_is_cycle_member=False,
            expected_is_root=False,
            expected_is_target=True,
        )

        intermediate_nodes = self._find_nodes_by_name(result_nodes, "intermediate")
        intermediate_node = one(intermediate_nodes)
        self._assert_node_properties(
            node=intermediate_node,
            expected_name="intermediate",
            expected_object_id=500,
            expected_is_cycle_member=False,
            expected_is_root=False,
            expected_is_target=False,
        )

        # Verify graph edges (remember: edges are reversed in the printable graph)
        # Original: root -> node1, node2 (root has 2 referents)
        #           node1 -> target1, target2, intermediate (node1 has 3 referents)
        #           node2 -> intermediate
        # Reversed: target1 -> node1 (original), target2 -> node1 (duplicate1),
        #           intermediate -> node1 (duplicate2), node2, node1 -> root (original),
        #           node2 -> root (duplicate)

        # Verify successors based on the debug output
        self._verify_node_successors(result, target1_node, ["node1"])
        self._verify_node_successors(result, target2_node, ["node1"])
        self._verify_node_successors(result, intermediate_node, ["node2", "node1"])
        self._verify_node_successors(result, original_node1, ["root"])
        self._verify_node_successors(result, node2_node, ["root"])

        # All root nodes and cycle member node1s should have no successors
        self._verify_node_successors(result, original_root, [])
        self._verify_node_successors(result, duplicate_root, [])
        for duplicate_node1 in duplicate_node1s:
            self._verify_node_successors(result, duplicate_node1, [])

        # Verify predecessors based on the debug output
        self._verify_node_predecessors(result, original_root, ["node1"])
        self._verify_node_predecessors(result, duplicate_root, ["node2"])
        self._verify_node_predecessors(result, original_node1, ["target1"])

        # Find which duplicate node1 is pointed to by which node
        duplicate_node1_with_target2_pred = None
        duplicate_node1_with_intermediate_pred = None
        for dup_node1 in duplicate_node1s:
            predecessors = [pred.name for pred in result.predecessors(dup_node1)]
            if "target2" in predecessors:
                duplicate_node1_with_target2_pred = dup_node1
            elif "intermediate" in predecessors:
                duplicate_node1_with_intermediate_pred = dup_node1

        assert duplicate_node1_with_target2_pred is not None
        assert duplicate_node1_with_intermediate_pred is not None

        self._verify_node_predecessors(result, duplicate_node1_with_target2_pred, ["target2"])
        self._verify_node_predecessors(
            result, duplicate_node1_with_intermediate_pred, ["intermediate"]
        )
        self._verify_node_predecessors(result, node2_node, ["intermediate"])

        # Targets and intermediate should have no predecessors
        self._verify_node_predecessors(result, target1_node, [])
        self._verify_node_predecessors(result, target2_node, [])
        self._verify_node_predecessors(result, intermediate_node, [])

    def test_complex_graph_with_multiple_characteristics(self) -> None:
        """
        Test convert_graph_nodes_to_printable with a complex graph having multiple targets,
        roots, and cycle members.

        Graph shape: Complex graph combining multiple features.
        root1 -> shared_node -> target1
        root2 -> shared_node (same object)
        root3 -> target2, intermediate (root3 has multiple referent_ids)
        """
        # given
        nodes = [
            ReferrerNode(id=0, name="root1", object_id=100, referent_ids=[3]),
            ReferrerNode(id=1, name="root2", object_id=200, referent_ids=[3]),
            ReferrerNode(
                id=2, name="root3", object_id=250, referent_ids=[5, 6]
            ),  # Multiple referent_ids
            ReferrerNode(id=3, name="shared_node", object_id=300, referent_ids=[4]),
            ReferrerNode(id=4, name="target1", object_id=400, referent_ids=[]),
            ReferrerNode(id=5, name="target2", object_id=500, referent_ids=[]),
            ReferrerNode(id=6, name="intermediate", object_id=600, referent_ids=[]),
        ]
        referrer_graph = ReferrerGraph(target_ids={4, 5}, root_ids={0, 1, 2}, graph_nodes=nodes)

        # when
        result = utils.convert_graph_nodes_to_printable(referrer_graph)

        # then
        result_nodes = list(result.nodes())

        # Should have 8 nodes: original 7 + 1 duplicate for root3's second referent_id
        assert len(result_nodes) == 8

        # Verify node type counts
        self._verify_node_counts_by_type(
            nodes=result_nodes,
            expected_targets=2,  # target1, target2
            expected_roots=4,  # root1, root2, root3 (original), root3 (duplicate)
            expected_cycle_members=1,  # duplicate root3
            expected_regular=2,  # shared_node, intermediate
        )

        # Verify individual nodes
        root1_nodes = self._find_nodes_by_name(result_nodes, "root1")
        root1_node = one(root1_nodes)
        self._assert_node_properties(
            node=root1_node,
            expected_name="root1",
            expected_object_id=100,
            expected_is_cycle_member=False,
            expected_is_root=True,
            expected_is_target=False,
        )

        root2_nodes = self._find_nodes_by_name(result_nodes, "root2")
        root2_node = one(root2_nodes)
        self._assert_node_properties(
            node=root2_node,
            expected_name="root2",
            expected_object_id=200,
            expected_is_cycle_member=False,
            expected_is_root=True,
            expected_is_target=False,
        )

        # root3 should have 2 instances: original + duplicate (cycle member)
        root3_nodes = self._find_nodes_by_name(result_nodes, "root3")
        assert len(root3_nodes) == 2

        original_root3 = next(node for node in root3_nodes if not node.is_cycle_member)
        duplicate_root3 = next(node for node in root3_nodes if node.is_cycle_member)

        self._assert_node_properties(
            node=original_root3,
            expected_name="root3",
            expected_object_id=250,
            expected_is_cycle_member=False,
            expected_is_root=True,
            expected_is_target=False,
        )

        self._assert_node_properties(
            node=duplicate_root3,
            expected_name="root3",
            expected_object_id=250,
            expected_is_cycle_member=True,
            expected_is_root=True,
            expected_is_target=False,
        )

        shared_node_nodes = self._find_nodes_by_name(result_nodes, "shared_node")
        shared_node = one(shared_node_nodes)
        self._assert_node_properties(
            node=shared_node,
            expected_name="shared_node",
            expected_object_id=300,
            expected_is_cycle_member=False,
            expected_is_root=False,
            expected_is_target=False,
        )

        target1_nodes = self._find_nodes_by_name(result_nodes, "target1")
        target1_node = one(target1_nodes)
        self._assert_node_properties(
            node=target1_node,
            expected_name="target1",
            expected_object_id=400,
            expected_is_cycle_member=False,
            expected_is_root=False,
            expected_is_target=True,
        )

        target2_nodes = self._find_nodes_by_name(result_nodes, "target2")
        target2_node = one(target2_nodes)
        self._assert_node_properties(
            node=target2_node,
            expected_name="target2",
            expected_object_id=500,
            expected_is_cycle_member=False,
            expected_is_root=False,
            expected_is_target=True,
        )

        intermediate_nodes = self._find_nodes_by_name(result_nodes, "intermediate")
        intermediate_node = one(intermediate_nodes)
        self._assert_node_properties(
            node=intermediate_node,
            expected_name="intermediate",
            expected_object_id=600,
            expected_is_cycle_member=False,
            expected_is_root=False,
            expected_is_target=False,
        )

        # Verify graph edges (remember: edges are reversed in the printable graph)
        # Original: root1 -> shared_node -> target1, root2 -> shared_node -> target1,
        #           root3 -> target2, intermediate (root3 has multiple referent_ids)
        # Reversed: target1 -> shared_node -> root1, root2
        #           target2 -> root3 (original), intermediate -> root3 (duplicate)
        self._verify_node_successors(result, target1_node, ["shared_node"])
        self._verify_node_successors(result, target2_node, ["root3"])
        self._verify_node_successors(result, shared_node, ["root1", "root2"])
        self._verify_node_successors(result, intermediate_node, ["root3"])

        # All root nodes should have no successors
        self._verify_node_successors(result, root1_node, [])
        self._verify_node_successors(result, root2_node, [])
        self._verify_node_successors(result, original_root3, [])
        self._verify_node_successors(result, duplicate_root3, [])

        # Verify predecessors
        self._verify_node_predecessors(result, root1_node, ["shared_node"])
        self._verify_node_predecessors(result, root2_node, ["shared_node"])
        self._verify_node_predecessors(result, original_root3, ["target2"])
        self._verify_node_predecessors(result, duplicate_root3, ["intermediate"])
        self._verify_node_predecessors(result, shared_node, ["target1"])
        self._verify_node_predecessors(result, target1_node, [])
        self._verify_node_predecessors(result, target2_node, [])
        self._verify_node_predecessors(result, intermediate_node, [])

    def _find_nodes_by_object_id(
        self, nodes: List[PrintableReferrerNode], object_id: int
    ) -> List[PrintableReferrerNode]:
        """
        Helper method to find all nodes with a specific object_id.
        """
        return [node for node in nodes if node.object_id == object_id]

    def _find_nodes_by_name(
        self, nodes: List[PrintableReferrerNode], name: str
    ) -> List[PrintableReferrerNode]:
        """
        Helper method to find all nodes with a specific name.
        """
        return [node for node in nodes if node.name == name]

    def _assert_node_properties(
        self,
        node: PrintableReferrerNode,
        expected_name: str,
        expected_object_id: int,
        expected_is_cycle_member: bool,
        expected_is_root: bool,
        expected_is_target: bool,
    ) -> None:
        """
        Helper method to assert all properties of a PrintableReferrerNode.
        """
        assert node.name == expected_name
        assert node.object_id == expected_object_id
        assert node.is_cycle_member == expected_is_cycle_member
        assert node.is_root == expected_is_root
        assert node.is_target == expected_is_target

    def _verify_node_counts_by_type(
        self,
        nodes: List[PrintableReferrerNode],
        expected_targets: int,
        expected_roots: int,
        expected_cycle_members: int,
        expected_regular: int,
    ) -> None:
        """
        Helper method to verify counts of different node types.
        """
        target_nodes = [node for node in nodes if node.is_target]
        root_nodes = [node for node in nodes if node.is_root]
        cycle_member_nodes = [node for node in nodes if node.is_cycle_member]
        regular_nodes = [
            node for node in nodes if not (node.is_target or node.is_root or node.is_cycle_member)
        ]

        assert len(target_nodes) == expected_targets
        assert len(root_nodes) == expected_roots
        assert len(cycle_member_nodes) == expected_cycle_members
        assert len(regular_nodes) == expected_regular

    def _verify_edge_exists(
        self, graph: nx.DiGraph, from_node: PrintableReferrerNode, to_node: PrintableReferrerNode
    ) -> None:
        """
        Helper method to verify that an edge exists between two nodes.
        """
        successors = list(graph.successors(from_node))
        assert to_node in successors, (
            f"Expected edge from {from_node.name} to {to_node.name} not found"
        )

    def _verify_node_successors(
        self, graph: nx.DiGraph, node: PrintableReferrerNode, expected_successor_names: List[str]
    ) -> None:
        """
        Helper method to verify that a node has exactly the expected successors.
        """
        successors = list(graph.successors(node))
        actual_successor_names = [succ.name for succ in successors]
        actual_successor_names.sort()
        expected_successor_names.sort()
        assert actual_successor_names == expected_successor_names, (
            f"Node {node.name} expected successors {expected_successor_names}, "
            f"but got {actual_successor_names}"
        )

    def _verify_node_predecessors(
        self, graph: nx.DiGraph, node: PrintableReferrerNode, expected_predecessor_names: List[str]
    ) -> None:
        """
        Helper method to verify that a node has exactly the expected predecessors.
        """
        predecessors = list(graph.predecessors(node))
        actual_predecessor_names = [pred.name for pred in predecessors]
        actual_predecessor_names.sort()
        expected_predecessor_names.sort()
        assert actual_predecessor_names == expected_predecessor_names, (
            f"Node {node.name} expected predecessors {expected_predecessor_names}, "
            f"but got {actual_predecessor_names}"
        )

    def _get_node_by_name_and_cycle_status(
        self, nodes: List[PrintableReferrerNode], name: str, is_cycle_member: bool
    ) -> PrintableReferrerNode:
        """
        Helper method to get a specific node by name and cycle member status.
        Useful when there are duplicate nodes with the same name.
        """
        matching_nodes = [
            node for node in nodes if node.name == name and node.is_cycle_member == is_cycle_member
        ]
        return one(matching_nodes)


class TestGetBudget:
    """
    Tests for the `get_budget` function.
    """

    @pytest.mark.parametrize(
        "object_type_counts,total_budget,expected_budget",
        [
            pytest.param(
                {},
                10,
                {},
                id="Empty types dictionary",
            ),
            pytest.param(
                {str: 1, int: 1, list: 1, dict: 1, tuple: 1},
                5,
                {str: 1, int: 1, list: 1, dict: 1, tuple: 1},
                id="Types equal to budget",
            ),
            pytest.param(
                {str: 1, int: 1, list: 1, dict: 1, tuple: 1, set: 1},
                5,
                {str: 1, int: 1, list: 1, dict: 1, tuple: 1, set: 1},
                id="More types than budget",
            ),
            pytest.param(
                {str: 1},
                10,
                {str: 10},
                id="Single type with budget larger than needed",
            ),
            pytest.param(
                {str: 1, list: 1},
                10,
                {str: 5, list: 5},
                id="Multiple types with budget larger than needed",
            ),
            pytest.param(
                {str: 20, int: 4},
                10,
                {str: 8, int: 2},
                id="Two types with proportional distribution",
            ),
            pytest.param(
                {str: 100, int: 50, list: 25},
                20,
                {str: 11, int: 6, list: 3},
                id="Three types with proportional distribution",
            ),
            pytest.param(
                {str: 10, int: 10},
                10,
                {str: 5, int: 5},
                id="Two types with equal counts",
            ),
            pytest.param(
                {str: 1000, int: 1},
                10,
                {str: 9, int: 1},
                id="Highly skewed distribution",
            ),
        ],
    )
    def test_get_object_type_budgets(
        self,
        object_type_counts: dict[type, int],
        total_budget: int,
        expected_budget: dict[type, int],
    ) -> None:
        """
        Tests that get_budget calculates the correct budget for object types.
        """
        budget = utils.get_object_type_budgets(
            object_type_counts=object_type_counts, total_budget=total_budget
        )
        assert budget == expected_budget



"""
Helper functions for integration tests.
"""

from __future__ import annotations

from pathlib import Path

from memalot.base import ReferrerGraph
from memalot.reports import (
    FileReportReader,
    FullReport,
    ObjectDetails,
    ReportIteration,
    TypeSummary,
)
from tests.utils_for_testing import one


def has_leak_iterations(
    reader: FileReportReader,
    expected_leak_iterations: int,
) -> bool:
    """
    Checks if the specified number of report iterations with leaks has been generated.
    """
    summaries = reader.get_report_summaries()
    if len(summaries) > 0:
        assert len(summaries) == 1, f"Expected exactly 1 summary, got {len(summaries)}"
        summary = summaries[0]
        if summary.iteration_count >= expected_leak_iterations:
            full_report_check = reader.get_full_report(
                report_id=summary.metadata.report_id, num_iterations=100
            )
            iterations_with_leaks = sum(
                1
                for iteration in full_report_check.iterations
                if len(iteration.leak_summary.type_summaries) > 0
            )
            if iterations_with_leaks >= expected_leak_iterations:
                return True
    return False


def read_single_report(tmp_path: Path) -> FullReport:
    """
    Reads a single report from the given directory.
    """
    reader = FileReportReader(report_directory=tmp_path)
    summaries = reader.get_report_summaries()
    assert len(summaries) == 1

    summary = one(summaries)
    metadata = summary.metadata
    full_report = reader.get_full_report(report_id=metadata.report_id, num_iterations=100)
    return full_report


def assert_iteration_count(full_report: FullReport, expected_count: int) -> None:
    """
    Asserts that the report has the expected number of iterations.
    """
    assert full_report.summary.iteration_count == expected_count
    assert len(full_report.iterations) == expected_count


def assert_warmup_iteration(iteration: ReportIteration) -> None:
    """
    Asserts that the given iteration is a warmup iteration with no leaks.
    """
    assert len(iteration.leak_summary.type_summaries) == 0
    assert len(list(iteration.object_details_list)) == 0


def assert_type_summaries(
    type_summaries: list[TypeSummary], expected_types: set[str], expected_counts: set[int]
) -> None:
    """
    Asserts that the type summaries contain the expected types and counts.
    """
    actual_types = {summary.object_type for summary in type_summaries}
    assert actual_types == expected_types, f"Expected types {expected_types}, got {actual_types}"
    actual_counts = {summary.count for summary in type_summaries}
    assert actual_counts == expected_counts, (
        f"Expected counts {expected_counts}, got {actual_counts}"
    )


def assert_object_details_types(
    object_details_list: list[ObjectDetails], expected_types: set[str]
) -> None:
    """
    Asserts that the object details list contains objects of the expected types.
    """
    object_details_types = {
        object_details.object_type_name for object_details in object_details_list
    }
    assert object_details_types == expected_types


def filter_object_details_by_type(
    object_details_list: list[ObjectDetails], object_type: str
) -> list[ObjectDetails]:
    """
    Filters the object details list to include only objects of the specified type.
    """
    return [
        object_details
        for object_details in object_details_list
        if object_details.object_type_name == object_type
    ]


def assert_referrer_graph_target_names(
    referrer_graph: ReferrerGraph, expected_target_names: set[str]
) -> None:
    """
    Asserts that the referrer graph has the expected target node names.
    """
    actual_target_names = {
        node.name for node in referrer_graph.graph_nodes if node.id in referrer_graph.target_ids
    }
    assert actual_target_names == expected_target_names, (
        f"Expected target names: {expected_target_names}, "
        f"Actual target names: {actual_target_names}"
    )


def assert_referrers_checked(object_details_list: list[ObjectDetails]) -> None:
    """
    Asserts that referrers have been checked for all object details.
    """
    for details in object_details_list:
        assert details.referrers_checked is True
        assert details.referrer_graph is not None


def assert_all_referrer_graphs_have_target_names(
    object_details_list: list[ObjectDetails], expected_target_names: set[str]
) -> None:
    """
    Asserts that all object details have referrer graphs with the expected target names.

    This helper verifies that:
    - All details have referrer graphs
    - Each referrer graph has the expected target names
    """
    for details in object_details_list:
        assert details.referrer_graph is not None
        assert_referrer_graph_target_names(
            referrer_graph=details.referrer_graph, expected_target_names=expected_target_names
        )


def assert_referrer_graphs_have_root_names(
    object_details_list: list[ObjectDetails], expected_root_names: set[str]
) -> None:
    """
    Asserts that the total set of root names in the referrer graphs for all object details
    is equal to the expected set.
    """
    all_root_names: set[str] = set()
    for details in object_details_list:
        assert details.referrer_graph is not None
        all_root_names.update(
            node.name
            for node in details.referrer_graph.graph_nodes
            if node.id in details.referrer_graph.root_ids
        )
    assert all_root_names == expected_root_names, (
        f"Expected root names: {expected_root_names}, Actual root names: {all_root_names}"
    )


def assert_object_type_details(
    object_details_list: list[ObjectDetails],
    object_type: str,
    expected_count: int,
    expected_target_names: set[str],
) -> list[ObjectDetails]:
    """
    Filters, counts, and verifies object details for a specific type.

    This helper performs the common pattern of:
    1. Filtering object details by type
    2. Asserting the expected count
    3. Verifying referrers are checked
    4. Verifying target names

    Returns the filtered list of object details for further assertions.
    """
    filtered_list = filter_object_details_by_type(
        object_details_list=object_details_list, object_type=object_type
    )
    assert len(filtered_list) == expected_count, (
        f"Expected {expected_count} {object_type} objects, found {len(filtered_list)}"
    )
    assert_referrers_checked(object_details_list=filtered_list)
    assert_all_referrer_graphs_have_target_names(
        object_details_list=filtered_list, expected_target_names=expected_target_names
    )
    return filtered_list



from abc import ABC, abstractmethod
from typing import Iterable

from rich.console import Console, RenderableType
from rich.console import NewLine as RichNewLine

from memalot.options import Options
from memalot.themes import DEFAULT_RICH_THEME


class Output(ABC):
    """
    An object that can be sent to an output sink. This may be the console, a file, etc.
    """

    @abstractmethod
    def to_renderables(self) -> Iterable[RenderableType]:  # pragma: no cover
        """
        Converts this object to a list of rich renderable objects.

        This is used when printing the object to the console, log files, etc.
        """
        pass


class RichOutput(Output):
    """
    An output object that returns the passed-in renderable object.
    """

    def __init__(self, renderable: RenderableType) -> None:
        self._renderable = renderable

    def to_renderables(self) -> Iterable[RenderableType]:
        return [self._renderable]


class NewLine(RichOutput):
    """
    A new line output object.
    """

    def __init__(self) -> None:
        super().__init__(RichNewLine())


class NewBlock(str):
    """
    A new block
    """


class OutputWriter(ABC):
    """
    A writer that sends objects to an output sink. This may be the console, a file, etc.
    """

    @abstractmethod
    def write(self, output: Output) -> None:  # pragma: no cover
        pass


class RichConsoleWriter(OutputWriter):
    def __init__(self, options: Options, console: Console) -> None:
        self._console = console
        self._options = options

    def write(self, output: Output) -> None:
        if self._options.output_func is not None:
            renderables = iter(output.to_renderables())
            exhausted = False
            while not exhausted:
                with self._console.capture() as captured:
                    try:
                        while True:
                            renderable = next(renderables)
                            if isinstance(renderable, NewBlock):
                                break
                            if type(renderable) is not RichNewLine and renderable != "\n":
                                self._console.print(
                                    RichNewLine(),
                                    soft_wrap=True,
                                    crop=False,
                                    overflow="ignore",
                                )
                                self._console.print(
                                    renderable,
                                    soft_wrap=True,
                                    crop=False,
                                    overflow="ignore",
                                )
                    except StopIteration:
                        exhausted = True
                        pass
                captured_str = captured.get()
                if captured_str.strip() != "":
                    self._options.output_func(captured_str)
        if self._options.output_func is None or self._options.tee_console:
            for renderable in output.to_renderables():
                self._console.print(renderable, soft_wrap=True, crop=False, overflow="ignore")


def get_output_writer(options: Options) -> OutputWriter:
    return RichConsoleWriter(
        options=options,
        console=Console(
            force_terminal=options.force_terminal,
            theme=DEFAULT_RICH_THEME,
            no_color=not options.color,
        ),
    )



from typing import Any, Dict, Optional
from unittest.mock import ANY, MagicMock

import pytest
from pytest_mock import MockerFixture

from memalot.api import create_leak_monitor, leak_monitor, start_leak_monitoring
from memalot.interface import LeakMonitor
from memalot.monitors import LeakMonitorImpl
from memalot.options import Options
from memalot.output import OutputWriter
from memalot.reports import ReportWriter
from tests.utils_for_testing import create_mock

_MODULE_UNDER_TEST = "memalot.api"


@pytest.fixture(name="mock_output_writer")
def _mock_output_writer() -> MagicMock:
    """
    Mock OutputWriter instance.
    """
    return create_mock(spec=OutputWriter)


@pytest.fixture(name="mock_report_writer")
def _mock_report_writer() -> MagicMock:
    """
    Mock ReportWriter instance.
    """
    return create_mock(spec=ReportWriter)


@pytest.fixture(name="mock_get_output_writer")
def _mock_get_output_writer(mocker: MockerFixture, mock_output_writer: MagicMock) -> MagicMock:
    """
    Mock get_output_writer function.
    """
    return mocker.patch(f"{_MODULE_UNDER_TEST}.get_output_writer", return_value=mock_output_writer)


@pytest.fixture(name="mock_get_report_writer")
def _mock_get_report_writer(mocker: MockerFixture, mock_report_writer: MagicMock) -> MagicMock:
    """
    Mock get_report_writer function.
    """
    return mocker.patch(f"{_MODULE_UNDER_TEST}.get_report_writer", return_value=mock_report_writer)


@pytest.fixture(name="mock_leak_monitor_thread")
def _mock_leak_monitor_thread(mocker: MockerFixture) -> MagicMock:
    """
    Mock LeakMonitorThread class.
    """
    return mocker.patch(f"{_MODULE_UNDER_TEST}.LeakMonitorThread")


@pytest.fixture(name="mock_leak_monitor")
def _mock_leak_monitor() -> MagicMock:
    """
    Mock LeakMonitor instance.
    """
    return create_mock(spec=LeakMonitor)


@pytest.fixture(name="mock_create_leak_monitor")
def _mock_create_leak_monitor(mocker: MockerFixture, mock_leak_monitor: MagicMock) -> MagicMock:
    """
    Mock _create_leak_monitor function.
    """
    return mocker.patch(f"{_MODULE_UNDER_TEST}.create_leak_monitor", return_value=mock_leak_monitor)


class TestStart:
    """
    Tests for the start function.
    """

    def test_start(
        self,
        mock_output_writer: MagicMock,
        mock_get_output_writer: MagicMock,
        mock_report_writer: MagicMock,
        mock_get_report_writer: MagicMock,
        mock_leak_monitor_thread: MagicMock,
    ) -> None:
        """
        Test that the start function starts the leak monitor.
        """
        expected_options = Options()

        start_leak_monitoring(max_object_lifetime=1.0)

        mock_get_output_writer.assert_called_once_with(options=expected_options)
        mock_get_report_writer.assert_called_once_with(options=expected_options)
        mock_leak_monitor_thread.assert_called_once_with(
            max_object_lifetime=1.0,
            warmup_time=1.0,
            writer=mock_output_writer,
            report_writer=mock_report_writer,
            options=expected_options,
            all_objects_manager=ANY,
            new_objects_manager=ANY,
            object_getter=ANY,
        )

    @pytest.mark.parametrize(
        "max_object_lifetime,warmup_time,expected_warmup_time",
        [
            (1.0, None, 1.0),  # warmup_time defaults to max_object_lifetime
            (2.0, 3.0, 3.0),  # explicit warmup_time
        ],
    )
    def test_start__with_max_object_lifetime_and_warmup_time(
        self,
        max_object_lifetime: float,
        warmup_time: Optional[float],
        expected_warmup_time: float,
        mock_output_writer: MagicMock,
        mock_get_output_writer: MagicMock,
        mock_report_writer: MagicMock,
        mock_get_report_writer: MagicMock,
        mock_leak_monitor_thread: MagicMock,
    ) -> None:
        """
        Test that the start function starts the leak monitor with different parameter values.
        """
        expected_options = Options()

        if warmup_time is None:
            start_leak_monitoring(max_object_lifetime=max_object_lifetime)
        else:
            start_leak_monitoring(max_object_lifetime=max_object_lifetime, warmup_time=warmup_time)

        mock_get_output_writer.assert_called_once_with(options=expected_options)
        mock_get_report_writer.assert_called_once_with(options=expected_options)
        mock_leak_monitor_thread.assert_called_once_with(
            max_object_lifetime=max_object_lifetime,
            warmup_time=expected_warmup_time,
            writer=mock_output_writer,
            report_writer=mock_report_writer,
            options=expected_options,
            all_objects_manager=ANY,
            new_objects_manager=ANY,
            object_getter=ANY,
        )

    @pytest.mark.parametrize(
        "kwargs,expected_options",
        [
            (
                {"force_terminal": True, "color": False},
                Options(force_terminal=True, color=False),
            ),
            (
                {"max_types_in_leak_summary": 100, "check_referrers": False},
                Options(
                    max_types_in_leak_summary=100,
                    check_referrers=False,
                ),
            ),
            (
                {"max_object_details": 50, "referrers_max_depth": 25},
                Options(max_object_details=50, referrers_max_depth=25),
            ),
        ],
    )
    def test_start__with_options_passed(
        self,
        kwargs: Dict[str, Any],
        expected_options: Options,
        mock_output_writer: MagicMock,
        mock_get_output_writer: MagicMock,
        mock_report_writer: MagicMock,
        mock_get_report_writer: MagicMock,
        mock_leak_monitor_thread: MagicMock,
    ) -> None:
        """
        Test that the start function passes options correctly to the Options class.
        """
        if "force_terminal" in kwargs and "color" in kwargs:
            start_leak_monitoring(
                max_object_lifetime=1.0,
                force_terminal=kwargs["force_terminal"],
                color=kwargs["color"],
            )
        elif "max_types_in_leak_summary" in kwargs and "check_referrers" in kwargs:
            start_leak_monitoring(
                max_object_lifetime=1.0,
                max_types_in_leak_summary=kwargs["max_types_in_leak_summary"],
                check_referrers=kwargs["check_referrers"],
            )
        elif "max_object_details" in kwargs and "referrers_max_depth" in kwargs:
            start_leak_monitoring(
                max_object_lifetime=1.0,
                max_object_details=kwargs["max_object_details"],
                referrers_max_depth=kwargs["referrers_max_depth"],
            )

        mock_get_output_writer.assert_called_once_with(options=expected_options)
        mock_get_report_writer.assert_called_once_with(options=expected_options)
        mock_leak_monitor_thread.assert_called_once_with(
            max_object_lifetime=1.0,
            warmup_time=1.0,
            writer=mock_output_writer,
            report_writer=mock_report_writer,
            options=expected_options,
            all_objects_manager=ANY,
            new_objects_manager=ANY,
            object_getter=ANY,
        )


class TestLeakMonitorDecorator:
    """
    Tests for the leak_monitor decorator.
    """

    def test_leak_monitor_decorator(
        self,
        mock_leak_monitor: MagicMock,
        mock_create_leak_monitor: MagicMock,
    ) -> None:
        """
        Test that the leak_monitor decorator works with default parameters.
        """

        @leak_monitor()
        def test_function(x: int, y: int) -> int:
            return x + y

        result = test_function(1, 2)

        assert result == 3
        self._check_create_leak_monitor_call_default(mock_create_leak_monitor)
        mock_leak_monitor.__enter__.assert_called_once()
        mock_leak_monitor.__exit__.assert_called_once()

    def test_leak_monitor_decorator__without_parentheses(
        self,
        mock_leak_monitor: MagicMock,
        mock_create_leak_monitor: MagicMock,
    ) -> None:
        """
        Test that the leak_monitor decorator works without parentheses.
        """

        @leak_monitor
        def test_function(x: int, y: int) -> int:
            return x + y

        result = test_function(1, 2)

        assert result == 3
        self._check_create_leak_monitor_call_default(mock_create_leak_monitor)
        mock_leak_monitor.__enter__.assert_called_once()
        mock_leak_monitor.__exit__.assert_called_once()

    @pytest.mark.parametrize(
        "warmup_calls,max_object_age_calls",
        [
            (1, 1),  # default values
            (2, 3),  # custom values
        ],
    )
    def test_leak_monitor_decorator__with_different_call_parameters(
        self,
        warmup_calls: int,
        max_object_age_calls: int,
        mock_leak_monitor: MagicMock,
        mock_create_leak_monitor: MagicMock,
    ) -> None:
        """
        Test that the leak_monitor decorator works with different parameter combinations.
        """

        @leak_monitor(warmup_calls=warmup_calls, min_object_age_calls=max_object_age_calls)
        def test_function(value: int) -> int:
            return value * 2

        result = test_function(10)

        assert result == 20
        mock_create_leak_monitor.assert_called_once_with(
            function_name="test_function",
            warmup_calls=warmup_calls,
            min_object_age_calls=max_object_age_calls,
            included_type_names=frozenset(),
            excluded_type_names=frozenset(),
            max_types_in_leak_summary=500,
            compute_size_in_leak_summary=False,
            max_untracked_search_depth=3,
            check_referrers=True,
            max_object_details=30,
            referrers_max_depth=50,
            referrers_search_timeout=300.0,
            single_object_referrer_limit=100,
            referrers_module_prefixes=None,
            referrers_max_untracked_search_depth=30,
            save_reports=True,
            report_directory=None,
            str_func=None,
            str_max_length=100,
            force_terminal=None,
            output_func=None,
            tee_console=False,
            color=True,
        )
        mock_leak_monitor.__enter__.assert_called_once()
        mock_leak_monitor.__exit__.assert_called_once()

    @pytest.mark.parametrize(
        "kwargs,expected_kwargs",
        [
            (
                {"force_terminal": True, "color": False},
                {"force_terminal": True, "color": False},
            ),
            (
                {"max_types_in_leak_summary": 100, "check_referrers": False},
                {"max_types_in_leak_summary": 100, "check_referrers": False},
            ),
            (
                {"max_object_details": 50, "referrers_max_depth": 25},
                {"max_object_details": 50, "referrers_max_depth": 25},
            ),
        ],
    )
    def test_leak_monitor_decorator__with_options_kwargs(
        self,
        kwargs: Dict[str, Any],
        expected_kwargs: Dict[str, Any],
        mock_leak_monitor: MagicMock,
        mock_create_leak_monitor: MagicMock,
    ) -> None:
        """
        Test that the leak_monitor decorator passes options kwargs correctly.
        """

        if "force_terminal" in kwargs and "color" in kwargs:

            @leak_monitor(
                warmup_calls=2,
                min_object_age_calls=3,
                force_terminal=kwargs["force_terminal"],
                color=kwargs["color"],
            )
            def test_function() -> str:
                return "success"

        elif "max_types_in_leak_summary" in kwargs and "check_referrers" in kwargs:

            @leak_monitor(
                warmup_calls=2,
                min_object_age_calls=3,
                max_types_in_leak_summary=kwargs["max_types_in_leak_summary"],
                check_referrers=kwargs["check_referrers"],
            )
            def test_function() -> str:
                return "success"

        elif "max_object_details" in kwargs and "referrers_max_depth" in kwargs:

            @leak_monitor(
                warmup_calls=2,
                min_object_age_calls=3,
                max_object_details=kwargs["max_object_details"],
                referrers_max_depth=kwargs["referrers_max_depth"],
            )
            def test_function() -> str:
                return "success"

        else:
            raise ValueError(f"Invalid kwargs: {kwargs}")

        result = test_function()

        assert result == "success"
        # Build the full expected call with all defaults
        expected_call_kwargs = {
            "function_name": "test_function",
            "warmup_calls": 2,
            "min_object_age_calls": 3,
            "included_type_names": frozenset(),
            "excluded_type_names": frozenset(),
            "max_types_in_leak_summary": 500,
            "compute_size_in_leak_summary": False,
            "max_untracked_search_depth": 3,
            "check_referrers": True,
            "max_object_details": 30,
            "referrers_max_depth": 50,
            "referrers_search_timeout": 300.0,
            "single_object_referrer_limit": 100,
            "referrers_module_prefixes": None,
            "referrers_max_untracked_search_depth": 30,
            "save_reports": True,
            "report_directory": None,
            "str_func": None,
            "str_max_length": 100,
            "force_terminal": None,
            "output_func": None,
            "tee_console": False,
            "color": True,
        }
        # Override with the specific expected kwargs
        expected_call_kwargs.update(expected_kwargs)
        mock_create_leak_monitor.assert_called_once_with(**expected_call_kwargs)
        mock_leak_monitor.__enter__.assert_called_once()
        mock_leak_monitor.__exit__.assert_called_once()

    def test_leak_monitor_decorator__exception_handling(
        self,
        mock_leak_monitor: MagicMock,
        mock_create_leak_monitor: MagicMock,
    ) -> None:
        """
        Test that the decorator properly handles exceptions from the decorated function.
        """

        @leak_monitor()
        def test_function() -> None:
            raise ValueError("test exception")

        with pytest.raises(ValueError, match="test exception"):
            test_function()

        self._check_create_leak_monitor_call_default(mock_create_leak_monitor)
        mock_leak_monitor.__enter__.assert_called_once()
        mock_leak_monitor.__exit__.assert_called_once()

    def _check_create_leak_monitor_call_default(self, mock_create_leak_monitor: MagicMock) -> None:
        mock_create_leak_monitor.assert_called_once_with(
            function_name="test_function",
            warmup_calls=1,
            min_object_age_calls=1,
            included_type_names=frozenset(),
            excluded_type_names=frozenset(),
            max_types_in_leak_summary=500,
            compute_size_in_leak_summary=False,
            max_untracked_search_depth=3,
            check_referrers=True,
            max_object_details=30,
            referrers_max_depth=50,
            referrers_search_timeout=300.0,
            single_object_referrer_limit=100,
            referrers_module_prefixes=None,
            referrers_max_untracked_search_depth=30,
            save_reports=True,
            report_directory=None,
            str_func=None,
            str_max_length=100,
            force_terminal=None,
            output_func=None,
            tee_console=False,
            color=True,
        )


class TestCreateLeakMonitor:
    """
    Tests for the _create_leak_monitor function.
    """

    def test_create_leak_monitor(
        self,
        mock_output_writer: MagicMock,
        mock_get_output_writer: MagicMock,
        mock_report_writer: MagicMock,
        mock_get_report_writer: MagicMock,
        mocker: MockerFixture,
    ) -> None:
        """
        Test that _create_leak_monitor creates a LeakMonitor with correct parameters.
        """
        mock_leak_monitor_impl = create_mock(spec=LeakMonitorImpl)
        mock_leak_monitor_impl_class = mocker.patch(
            f"{_MODULE_UNDER_TEST}.LeakMonitorImpl", return_value=mock_leak_monitor_impl
        )

        expected_options = Options(save_reports=True)

        result = create_leak_monitor(
            warmup_calls=5,
            min_object_age_calls=3,
        )

        assert result == mock_leak_monitor_impl
        mock_get_output_writer.assert_called_once_with(options=expected_options)
        mock_get_report_writer.assert_called_once_with(options=expected_options)
        mock_leak_monitor_impl_class.assert_called_once_with(
            writer=mock_output_writer,
            report_writer=mock_report_writer,
            warmup_calls=5,
            calls_per_report=3,
            options=expected_options,
            snapshot_manager=ANY,
            object_getter=ANY,
            function_name=None,
        )



"""
Performance test based on the barbell example.
"""

import random
from dataclasses import dataclass, field

import networkx as nx
import pytest
from pytest_benchmark.fixture import BenchmarkFixture

from memalot import leak_monitor


@dataclass
class Node:
    """
    A node in a graph.
    """

    value: int
    adjacent_nodes: list["Node"] = field(default_factory=list)


def create_graph(m1: int, m2: int) -> tuple[dict[str, Node], Node]:
    """
    Creates a barbell graph with m1 nodes in each complete graph and m2 nodes in the path.

    Returns a tuple of (cache, end_node).
    """
    graph = nx.barbell_graph(m1, m2)
    node_map = {node_id: Node(node_id) for node_id in graph.nodes()}
    for node_id, node_obj in node_map.items():
        for neighbor_id in graph.neighbors(node_id):
            neighbor_obj = node_map[neighbor_id]
            node_obj.adjacent_nodes.append(neighbor_obj)
    start_node = node_map[0]
    cache = {"start_node": start_node}
    end_node = node_map[max(node_map.keys())]
    return cache, end_node


@leak_monitor(
    check_referrers=True,
    min_object_age_calls=1,
    max_object_details=2,
    # Set output to nowhere
    output_func=lambda _: None,
)
def add_nodes_to_end(end_node: Node, num_new_nodes: int) -> None:
    """
    Adds new nodes to the end node.
    """
    for _ in range(num_new_nodes):
        new_node = Node(random.randint(1, 1000))
        end_node.adjacent_nodes.append(new_node)


@pytest.mark.performance
class TestBarbellPerformance:
    """
    Performance tests based on the barbell example.
    """

    def test_barbell(self, benchmark: BenchmarkFixture) -> None:
        """
        Test the performance of the barbell example with m1=30, m2=20, two iterations,
        and max_object_summaries=2.
        """

        def run_barbell() -> None:
            cache, end_node = create_graph(m1=30, m2=20)
            # Run for two iterations
            add_nodes_to_end(end_node, 3)
            add_nodes_to_end(end_node, 3)

        benchmark.pedantic(run_barbell, rounds=1)

        # Based on preliminary testing, the test should complete in well under 200 seconds
        # on average, even on Github's workers (this is about 4x what it takes on a fast laptop).
        # This assertion allows for some variability but catches very significant performance
        # regressions.
        assert benchmark.stats.stats.mean < 200.0, (
            f"Performance regression detected: mean time {benchmark.stats.stats.mean:.2f}s "
            f"exceeds 200s"
        )
